import numpy as np


def root_mean_squared_percentage_error(y_true, y_pred):
    """Compute RMSPE."""
    mask = y_true > 0.0
    y_true = y_true[mask]
    y_pred = y_pred[mask]
    score = (y_true - y_pred) / y_pred
    return np.sqrt(np.mean(score ** 2)) * 100



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))

# -- Project information -----------------------------------------------------

project = "glum"
copyright = "(C) 2020-2021 QuantCo Inc."
author = "QuantCo Inc."

extensions = [
    "sphinx.ext.napoleon",
    "altair.sphinxext.altairplot",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "nbsphinx",
    "sphinx.ext.mathjax",
]

autodoc_typehints = "description"

templates_path = ["_templates"]
exclude_patterns = [
    "modules.rst",
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "**.ipynb_checkpoints",
    "lasso.ipynb",
    "spatial-smoothing.ipynb",
    "benchmarks/*",
    "tutorials/rossman/explore_data.ipynb",
]
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_theme_options = {
    # this puts the entire table of contents structure in the sidebar.
    # unfortunately, it's not possible yet to have it expanded by default.
    # see: https://github.com/readthedocs/sphinx_rtd_theme/issues/455
    "collapse_navigation": False,
    "navigation_depth": 2,
}

# For the altairplot extension
altairplot_links = {"editor": True, "source": True, "export": True}



import numpy as np
import pandas as pd
import pytest

from glum._util import _align_df_categories


@pytest.fixture()
def df():
    return pd.DataFrame(
        {
            "x1": np.array([0, 1], dtype="int64"),
            "x2": np.array([0, 1], dtype="bool"),
            "x3": np.array([0, 1], dtype="float64"),
            "x4": ["0", "1"],
            "x5": ["a", "b"],
            "x6": pd.Categorical(["a", "b"]),
            "x7": pd.Categorical(["a", "b"], categories=["b", "a"]),
        }
    )


def test_align_df_categories_numeric(df):

    dtypes = {column: np.float64 for column in df}

    expected = pd.DataFrame(
        {
            "x1": np.array([0, 1], dtype="int64"),
            "x2": np.array([0, 1], dtype="bool"),
            "x3": np.array([0, 1], dtype="float64"),
            "x4": ["0", "1"],
            "x5": ["a", "b"],
            "x6": pd.Categorical(["a", "b"]),
            "x7": pd.Categorical(["a", "b"], categories=["b", "a"]),
        }
    )

    pd.testing.assert_frame_equal(_align_df_categories(df, dtypes), expected)


def test_align_df_categories_categorical(df):

    dtypes = {column: pd.CategoricalDtype(["a", "b"]) for column in df}

    expected = pd.DataFrame(
        {
            "x1": [np.nan, np.nan],
            "x2": [np.nan, np.nan],
            "x3": [np.nan, np.nan],
            "x4": [np.nan, np.nan],
            "x5": pd.Categorical(["a", "b"]),
            "x6": pd.Categorical(["a", "b"]),
            "x7": pd.Categorical(["a", "b"]),
        },
        dtype=pd.CategoricalDtype(["a", "b"]),
    )

    pd.testing.assert_frame_equal(_align_df_categories(df, dtypes), expected)


def test_align_df_categories_excess_columns(df):

    dtypes = {"x1": np.float64}

    expected = pd.DataFrame(
        {
            "x1": np.array([0, 1], dtype="int64"),
            "x2": np.array([0, 1], dtype="bool"),
            "x3": np.array([0, 1], dtype="float64"),
            "x4": ["0", "1"],
            "x5": ["a", "b"],
            "x6": pd.Categorical(["a", "b"]),
            "x7": pd.Categorical(["a", "b"], categories=["b", "a"]),
        }
    )

    pd.testing.assert_frame_equal(_align_df_categories(df, dtypes), expected)


def test_align_df_categories_missing_columns(df):

    dtypes = {"x0": np.float64}

    expected = pd.DataFrame(
        {
            "x1": np.array([0, 1], dtype="int64"),
            "x2": np.array([0, 1], dtype="bool"),
            "x3": np.array([0, 1], dtype="float64"),
            "x4": ["0", "1"],
            "x5": ["a", "b"],
            "x6": pd.Categorical(["a", "b"]),
            "x7": pd.Categorical(["a", "b"], categories=["b", "a"]),
        }
    )

    pd.testing.assert_frame_equal(_align_df_categories(df, dtypes), expected)


def test_align_df_categories_not_df():
    with pytest.raises(TypeError):
        _align_df_categories(np.array([[0], [1]]), {"x0": np.float64})



from typing import Any, Dict, Union

import numpy as np
from scipy import sparse as sps


def zeros_bench(
    dat: Dict[str, Union[np.ndarray, sps.spmatrix]], *, cv: bool = False, **kwargs
) -> Dict[str, Any]:
    """
    Run a "benchmark" of how long it takes to return all zero coefficients.

    Parameters
    ----------
    dat
    cv
    kwargs
    """
    result = {
        "runtime": 0,
        "model_obj": None,
        "intercept": 0,
        "coef": np.zeros(dat["X"].shape[1]),
        "n_iter": 1,
    }
    if cv:
        result["alpha"] = 0
    return result



# Authors: Christian Lorentzen <lorentzen.ch@gmail.com>
#
# License: BSD 3 clause
import copy
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pytest
import statsmodels.api as sm
import tabmat as tm
from numpy.testing import assert_allclose
from scipy import optimize, sparse
from sklearn.base import clone
from sklearn.datasets import make_classification, make_regression
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import ElasticNet, LogisticRegression, Ridge
from sklearn.metrics import mean_absolute_error
from sklearn.utils.estimator_checks import check_estimator

from glum import GeneralizedLinearRegressorCV
from glum._distribution import (
    BinomialDistribution,
    ExponentialDispersionModel,
    GammaDistribution,
    GeneralizedHyperbolicSecant,
    InverseGaussianDistribution,
    NormalDistribution,
    PoissonDistribution,
    TweedieDistribution,
    guess_intercept,
)
from glum._glm import GeneralizedLinearRegressor, _unstandardize, is_pos_semidef
from glum._link import IdentityLink, Link, LogitLink, LogLink

GLM_SOLVERS = ["irls-ls", "lbfgs", "irls-cd", "trust-constr"]

estimators = [
    (GeneralizedLinearRegressor, {}),
    (GeneralizedLinearRegressorCV, {"n_alphas": 2}),
]


def get_small_x_y(
    estimator: Union[GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
) -> Tuple[np.ndarray, np.ndarray]:
    if isinstance(estimator, GeneralizedLinearRegressor):
        n_rows = 1
    else:
        n_rows = 10
    x = np.ones((n_rows, 1), dtype=int)
    y = np.ones(n_rows) * 0.5
    return x, y


@pytest.fixture(scope="module")
def regression_data():
    X, y = make_regression(
        n_samples=107, n_features=10, n_informative=80, noise=0.5, random_state=2
    )
    return X, y


@pytest.fixture
def y():
    """Get values for y that are in range of all distributions."""
    return np.array([0.1, 0.5])


@pytest.fixture
def X():
    return np.array([[1], [2]])


@pytest.mark.parametrize("estimator, kwargs", estimators)
def test_sample_weights_validation(estimator, kwargs):
    """Test the raised errors in the validation of sample_weight."""
    # scalar value but not positive
    X, y = get_small_x_y(estimator)
    sample_weight: Any = 0
    glm = estimator(fit_intercept=False, **kwargs)
    with pytest.raises(ValueError, match="weights must be non-negative"):
        glm.fit(X, y, sample_weight)

    # Positive weights are accepted
    glm.fit(X, y, sample_weight=1)

    # 2d array
    sample_weight = [[0]]
    with pytest.raises(ValueError, match="must be 1D array or scalar"):
        glm.fit(X, y, sample_weight)

    # 1d but wrong length
    sample_weight = [1, 0]
    with pytest.raises(ValueError, match="weights must have the same length as y"):
        glm.fit(X, y, sample_weight)

    # 1d but only zeros (sum not greater than 0)
    sample_weight = [0, 0]
    X = [[0], [1]]
    y = [1, 2]
    with pytest.raises(ValueError, match="must have at least one positive element"):
        glm.fit(X, y, sample_weight)

    # 5. 1d but with a negative value
    sample_weight = [2, -1]
    with pytest.raises(ValueError, match="weights must be non-negative"):
        glm.fit(X, y, sample_weight)


@pytest.mark.parametrize("estimator, kwargs", estimators)
def test_offset_validation(estimator, kwargs):
    X, y = get_small_x_y(estimator)
    glm = estimator(fit_intercept=False, **kwargs)

    # Negatives are accepted (makes sense for log link)
    glm.fit(X, y, offset=-1)

    # Arrays of the right shape are accepted
    glm.fit(X, y, offset=y.copy())

    # 2d array
    with pytest.raises(ValueError, match="must be 1D array or scalar"):
        glm.fit(X, y, offset=np.zeros_like(X))

    # 1d but wrong length
    with pytest.raises(ValueError, match="must have the same length as y"):
        glm.fit(X, y, offset=[1, 0])


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
def test_tol_validation_errors(estimator):
    X, y = get_small_x_y(estimator)

    glm = estimator(gradient_tol=-0.1)
    with pytest.raises(ValueError, match="Tolerance for the gradient stopping"):
        glm.fit(X, y)

    glm = estimator(step_size_tol=-0.1)
    with pytest.raises(ValueError, match="Tolerance for the step-size stopping"):
        glm.fit(X, y)


@pytest.mark.parametrize("estimator, kwargs", estimators)
@pytest.mark.parametrize(
    "tol_kws",
    [
        {},
        {"step_size_tol": 1},
        {"step_size_tol": None},
        {"gradient_tol": 1},
        {"gradient_tol": 1, "step_size_tol": 1},
    ],
)
def test_tol_validation_no_error(estimator, kwargs, tol_kws):
    X, y = get_small_x_y(estimator)
    glm = estimator(**tol_kws, **kwargs)
    glm.fit(X, y)


@pytest.mark.parametrize("estimator, kwargs", estimators)
@pytest.mark.parametrize("solver", ["auto", "irls-cd", "trust-constr"])
@pytest.mark.parametrize("gradient_tol", [None, 1])
def test_gradient_tol_setting(estimator, kwargs, solver, gradient_tol):
    X, y = get_small_x_y(estimator)
    glm = estimator(solver=solver, gradient_tol=gradient_tol, **kwargs)
    glm.fit(X, y)

    if gradient_tol is None:
        if solver == "trust-constr":
            gradient_tol = 1e-8
        else:
            gradient_tol = 1e-4

    np.testing.assert_allclose(gradient_tol, glm._gradient_tol)


# TODO: something for CV regressor
@pytest.mark.parametrize(
    "f, fam",
    [
        ("gaussian", NormalDistribution()),
        ("normal", NormalDistribution()),
        ("poisson", PoissonDistribution()),
        ("gamma", GammaDistribution()),
        ("inverse.gaussian", InverseGaussianDistribution()),
        ("binomial", BinomialDistribution()),
    ],
)
def test_glm_family_argument(f, fam, y, X):
    """Test GLM family argument set as string."""
    glm = GeneralizedLinearRegressor(family=f, alpha=0).fit(X, y)
    assert isinstance(glm._family_instance, fam.__class__)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
def test_glm_family_argument_invalid_input(estimator):
    X, y = get_small_x_y(estimator)
    glm = estimator(family="not a family", fit_intercept=False)
    with pytest.raises(ValueError, match="family must be"):
        glm.fit(X, y)


@pytest.mark.parametrize("estimator, kwargs", estimators)
@pytest.mark.parametrize("family", ExponentialDispersionModel.__subclasses__())
def test_glm_family_argument_as_exponential_dispersion_model(estimator, kwargs, family):
    X, y = get_small_x_y(estimator)
    glm = estimator(family=family(), **kwargs)
    glm.fit(X, y)


@pytest.mark.parametrize(
    "link_func, link",
    [("identity", IdentityLink()), ("log", LogLink()), ("logit", LogitLink())],
)
def test_glm_link_argument(link_func, link, y, X):
    """Test GLM link argument set as string."""
    glm = GeneralizedLinearRegressor(family="normal", link=link_func).fit(X, y)
    assert isinstance(glm._link_instance, link.__class__)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
def test_glm_link_argument_invalid_input(estimator):
    X, y = get_small_x_y(estimator)
    glm = estimator(family="normal", link="not a link")
    with pytest.raises(ValueError, match="link must be"):
        glm.fit(X, y)


@pytest.mark.parametrize("alpha", ["not a number", -4.2])
def test_glm_alpha_argument(alpha, y, X):
    """Test GLM for invalid alpha argument."""
    glm = GeneralizedLinearRegressor(family="normal", alpha=alpha)
    with pytest.raises(ValueError, match="Penalty term must be a non-negative"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("l1_ratio", ["not a number", -4.2, 1.1])
def test_glm_l1_ratio_argument(estimator, l1_ratio):
    """Test GLM for invalid l1_ratio argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(family="normal", l1_ratio=l1_ratio)
    with pytest.raises(ValueError, match="l1_ratio must be a number in interval.*0, 1"):
        glm.fit(X, y)


def test_glm_ratio_argument_array():
    X, y = get_small_x_y(GeneralizedLinearRegressor)
    glm = GeneralizedLinearRegressor(family="normal", l1_ratio=[1])
    with pytest.raises(ValueError, match="l1_ratio must be a number in interval.*0, 1"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("P1", [["a string", "a string"], [1, [2]], [1, 2, 3], [-1]])
def test_glm_P1_argument(estimator, P1, y, X):
    """Test GLM for invalid P1 argument."""
    glm = estimator(P1=P1, l1_ratio=0.5, check_input=True)
    with pytest.raises((ValueError, TypeError)):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize(
    "P2", ["a string", [1, 2, 3], [[2, 3]], sparse.csr_matrix([1, 2, 3]), [-1]]
)
def test_glm_P2_argument(estimator, P2, y, X):
    """Test GLM for invalid P2 argument."""
    glm = estimator(P2=P2, check_input=True)
    with pytest.raises(ValueError):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
def test_glm_P2_positive_semidefinite(estimator):
    """Test GLM for a positive semi-definite P2 argument."""
    n_samples, n_features = 10, 2
    y = np.arange(n_samples)
    X = np.zeros((n_samples, n_features))

    # negative definite matrix
    P2 = np.array([[1, 2], [2, 1]])
    glm = estimator(P2=P2, fit_intercept=False, check_input=True)
    with pytest.raises(ValueError, match="P2 must be positive semi-definite"):
        glm.fit(X, y)

    P2 = sparse.csr_matrix(P2)
    glm = estimator(P2=P2, fit_intercept=False, check_input=True)
    with pytest.raises(ValueError, match="P2 must be positive semi-definite"):
        glm.fit(X, y)


def test_positive_semidefinite():
    """Test GLM for a positive semi-definite P2 argument."""
    # negative definite matrix
    P2 = np.array([[1, 2], [2, 1]])
    assert not is_pos_semidef(P2)

    P2 = sparse.csr_matrix(P2)
    assert not is_pos_semidef(P2)

    assert is_pos_semidef(np.eye(2))
    assert is_pos_semidef(sparse.eye(2))


def test_P1_P2_expansion_with_categoricals():
    rng = np.random.default_rng(42)
    X = pd.DataFrame(
        data={
            "dense": np.linspace(0, 10, 60),
            "cat": pd.Categorical(rng.integers(5, size=60)),
        }
    )
    y = rng.normal(size=60)

    mdl1 = GeneralizedLinearRegressor(
        l1_ratio=0.01,
        P1=[1, 2, 2, 2, 2, 2],
        P2=[2, 1, 1, 1, 1, 1],
    )
    mdl1.fit(X, y)

    mdl2 = GeneralizedLinearRegressor(
        l1_ratio=0.01,
        P1=[1, 2],
        P2=[2, 1],
    )
    mdl2.fit(X, y)
    np.testing.assert_allclose(mdl1.coef_, mdl2.coef_)

    mdl2 = GeneralizedLinearRegressor(
        l1_ratio=0.01, P1=[1, 2], P2=sparse.diags([2, 1, 1, 1, 1, 1])
    )
    mdl2.fit(X, y)
    np.testing.assert_allclose(mdl1.coef_, mdl2.coef_)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("fit_intercept", ["not bool", 1, 0, [True]])
def test_glm_fit_intercept_argument(estimator, fit_intercept):
    """Test GLM for invalid fit_intercept argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(fit_intercept=fit_intercept)
    with pytest.raises(TypeError, match="fit_intercept must be bool"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize(
    "solver, l1_ratio",
    [
        ("not a solver", 0),
        (1, 0),
        ([1], 0),
        ("irls-ls", 0.5),
        ("lbfgs", 0.5),
        ("trust-constr", 0.5),
    ],
)
def test_glm_solver_argument(estimator, solver, l1_ratio, y, X):
    """Test GLM for invalid solver argument."""
    glm = estimator(solver=solver, l1_ratio=l1_ratio)
    with pytest.raises(ValueError):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("max_iter", ["not a number", 0, -1, 5.5, [1]])
def test_glm_max_iter_argument(estimator, max_iter):
    """Test GLM for invalid max_iter argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(max_iter=max_iter)
    with pytest.raises(ValueError, match="must be a positive integer"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("tol_param", ["gradient_tol", "step_size_tol"])
@pytest.mark.parametrize("tol", ["not a number", 0, -1.0, [1e-3]])
def test_glm_tol_argument(estimator, tol_param, tol):
    """Test GLM for invalid tol argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(**{tol_param: tol})
    with pytest.raises(ValueError, match="stopping criteria must be positive"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("warm_start", ["not bool", 1, 0, [True]])
def test_glm_warm_start_argument(estimator, warm_start):
    """Test GLM for invalid warm_start argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(warm_start=warm_start)
    with pytest.raises(TypeError, match="warm_start must be bool"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize(
    "start_params", ["not a start_params", ["zero"], [0, 0, 0], [[0, 0]], ["a", "b"]]
)
def test_glm_start_params_argument(estimator, start_params, y, X):
    """Test GLM for invalid start_params argument."""
    glm = estimator(start_params=start_params)
    with pytest.raises(ValueError):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("selection", ["not a selection", 1, 0, ["cyclic"]])
def test_glm_selection_argument(estimator, selection):
    """Test GLM for invalid selection argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(selection=selection)
    with pytest.raises(ValueError, match="argument selection must be"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("random_state", ["a string", 0.5, [0]])
def test_glm_random_state_argument(estimator, random_state):
    """Test GLM for invalid random_state argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(random_state=random_state)
    with pytest.raises(ValueError, match="cannot be used to seed"):
        glm.fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("copy_X", ["not bool", 1, 0, [True]])
def test_glm_copy_X_argument_invalid(estimator, copy_X):
    """Test GLM for invalid copy_X arguments."""
    X, y = get_small_x_y(estimator)
    glm = estimator(copy_X=copy_X)
    with pytest.raises(TypeError, match="copy_X must be None or bool"):
        glm.fit(X, y)


def test_glm_copy_X_input_needs_conversion():
    y = np.array([1.0])
    # If X is of int dtype, it needs to be copied
    X = np.array([[1]])
    glm = GeneralizedLinearRegressor(copy_X=False)
    # should raise an error
    with pytest.raises(ValueError, match="copy_X"):
        glm.fit(X, y)
    # should be OK with copy_X = None or copy_X = True
    GeneralizedLinearRegressor(copy_X=None).fit(X, y)
    GeneralizedLinearRegressor(copy_X=True).fit(X, y)


@pytest.mark.parametrize(
    "estimator", [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV]
)
@pytest.mark.parametrize("check_input", ["not bool", 1, 0, [True]])
def test_glm_check_input_argument(estimator, check_input):
    """Test GLM for invalid check_input argument."""
    X, y = get_small_x_y(estimator)
    glm = estimator(check_input=check_input)
    with pytest.raises(TypeError, match="check_input must be bool"):
        glm.fit(X, y)


@pytest.mark.parametrize("solver", GLM_SOLVERS)
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("offset", [None, np.array([-0.1, 0, 0.1, 0, -0.2]), 0.1])
@pytest.mark.parametrize(
    "convert_x_fn",
    [
        np.asarray,
        sparse.csc_matrix,
        sparse.csr_matrix,
        tm.DenseMatrix,
        lambda x: tm.SparseMatrix(sparse.csc_matrix(x)),
        lambda x: tm.from_csc(sparse.csc_matrix(x)),
    ],
)
def test_glm_identity_regression(solver, fit_intercept, offset, convert_x_fn):
    """Test GLM regression with identity link on a simple dataset."""
    coef = [1.0, 2.0]
    X = np.array([[1, 1, 1, 1, 1], [0, 1, 2, 3, 4]]).T
    y = np.dot(X, coef) + (0 if offset is None else offset)
    glm = GeneralizedLinearRegressor(
        alpha=0,
        family="normal",
        link="identity",
        fit_intercept=fit_intercept,
        solver=solver,
        gradient_tol=1e-7,
    )
    if fit_intercept:
        X = X[:, 1:]

    X = convert_x_fn(X.astype(float))
    res = glm.fit(X, y, offset=offset)
    if fit_intercept:
        fit_coef = np.concatenate([[res.intercept_], res.coef_])
    else:
        fit_coef = res.coef_
    assert fit_coef.dtype.itemsize == X.dtype.itemsize
    assert_allclose(fit_coef, coef, rtol=1e-6)


@pytest.mark.parametrize("solver", GLM_SOLVERS)
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("full_report", [False, True])
@pytest.mark.parametrize("custom_columns", [None, ["objective_fct"]])
def test_get_diagnostics(
    solver, fit_intercept: bool, full_report: bool, custom_columns: Optional[List[str]]
):
    """Test GLM regression with identity link on a simple dataset."""
    X, y = get_small_x_y(GeneralizedLinearRegressor)

    glm = GeneralizedLinearRegressor(fit_intercept=fit_intercept, solver=solver)
    res = glm.fit(X, y)

    diagnostics = res.get_formatted_diagnostics(full_report, custom_columns)
    if solver in ("lbfgs", "trust-constr"):
        assert diagnostics == "solver does not report diagnostics"
    else:
        assert diagnostics.index.name == "n_iter"
        if custom_columns is not None:
            expected_columns = custom_columns
        elif full_report:
            expected_columns = [
                "convergence",
                "objective_fct",
                "L1(coef)",
                "L2(coef)",
                "L2(step)",
                "first_coef",
                "n_coef_updated",
                "n_active_cols",
                "n_active_rows",
                "n_cycles",
                "n_line_search",
                "iteration_runtime",
                "build_hessian_runtime",
                "inner_solver_runtime",
                "line_search_runtime",
                "quadratic_update_runtime",
                "intercept",
            ]
        else:
            expected_columns = [
                "convergence",
                "n_cycles",
                "iteration_runtime",
                "intercept",
            ]
        assert (diagnostics.columns == expected_columns).all()
        if "intercept" in expected_columns:
            null_intercept = diagnostics["intercept"].isnull()
            if fit_intercept:
                assert not null_intercept.any()
            else:
                assert null_intercept.all()


@pytest.mark.parametrize("solver", GLM_SOLVERS)
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("full_report", [False, True])
@pytest.mark.parametrize("custom_columns", [None, ["objective_fct"]])
def test_report_diagnostics(
    solver, fit_intercept: bool, full_report: bool, custom_columns: Optional[List[str]]
):
    """Test GLM regression with identity link on a simple dataset."""
    X, y = get_small_x_y(GeneralizedLinearRegressor)

    glm = GeneralizedLinearRegressor(fit_intercept=fit_intercept, solver=solver)
    res = glm.fit(X, y)

    # Make sure something prints
    import io
    from contextlib import redirect_stdout

    f = io.StringIO()
    with redirect_stdout(f):
        res.report_diagnostics(full_report, custom_columns)
    printed = f.getvalue()
    # Something should be printed
    assert len(printed) > 0


@pytest.mark.parametrize("solver", GLM_SOLVERS)
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("offset", [None, np.array([-0.1, 0, 0.1, 0, -0.2]), 0.1])
@pytest.mark.parametrize(
    "convert_x_fn",
    [
        np.asarray,
        sparse.csc_matrix,
        sparse.csr_matrix,
        tm.DenseMatrix,
        lambda x: tm.SparseMatrix(sparse.csc_matrix(x)),
        lambda x: tm.from_csc(sparse.csc_matrix(x)),
    ],
)
def test_x_not_modified_inplace(solver, fit_intercept, offset, convert_x_fn):
    coef = [1.0, 2.0]
    X = np.array([[1, 1, 1, 1, 1], [0, 1, 2, 3, 4]]).T
    y = np.dot(X, coef) + (0 if offset is None else offset)
    glm = GeneralizedLinearRegressor(
        alpha=0,
        family="normal",
        link="identity",
        fit_intercept=fit_intercept,
        solver=solver,
        gradient_tol=1e-7,
    )
    if fit_intercept:
        X = X[:, 1:]

    X = convert_x_fn(X.astype(float))

    X_before = copy.deepcopy(X)
    glm.fit(X, y, offset=offset)
    if isinstance(X, np.ndarray):
        np.testing.assert_almost_equal(X, X_before)
    else:
        np.testing.assert_almost_equal(X.A, X_before.A)


@pytest.mark.parametrize("solver", GLM_SOLVERS)
@pytest.mark.parametrize("offset", [None, np.array([-0.1, 0, 0.1, 0, -0.2]), 0.1])
@pytest.mark.parametrize(
    "convert_x_fn",
    [
        np.asarray,
        sparse.csc_matrix,
        sparse.csr_matrix,
        lambda x: tm.DenseMatrix(x.astype(float)),
        lambda x: tm.SparseMatrix(sparse.csc_matrix(x)),
        lambda x: tm.from_csc(sparse.csc_matrix(x).astype(float)),
        lambda x: tm.CategoricalMatrix(x.dot([0, 1])),
    ],
)
def test_glm_identity_regression_categorical_data(solver, offset, convert_x_fn):
    """Test GLM regression with identity link on a simple dataset."""
    coef = [1.0, 2.0]
    x_vec = np.array([1, 0, 0, 1, 0])
    x_mat = np.stack([x_vec, 1 - x_vec]).T
    y = np.dot(x_mat, coef) + (0 if offset is None else offset)

    glm = GeneralizedLinearRegressor(
        alpha=0,
        family="normal",
        link="identity",
        fit_intercept=False,
        solver=solver,
        gradient_tol=1e-7,
    )
    X = convert_x_fn(x_mat)
    np.testing.assert_almost_equal(X.A if hasattr(X, "A") else X, x_mat)
    res = glm.fit(X, y, offset=offset)

    assert_allclose(res.coef_, coef, rtol=1e-6)


@pytest.mark.parametrize(
    "family",
    [
        NormalDistribution(),
        PoissonDistribution(),
        GammaDistribution(),
        InverseGaussianDistribution(),
        TweedieDistribution(power=1.5),
        TweedieDistribution(power=4.5),
        GeneralizedHyperbolicSecant(),
    ],
)
@pytest.mark.parametrize(
    "solver, tol",
    [("irls-ls", 1e-6), ("lbfgs", 1e-7), ("irls-cd", 1e-7), ("trust-constr", 1e-7)],
)
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.parametrize("offset", [None, np.array([-0.1, 0, 0.1, 0, -0.2]), 0.1])
def test_glm_log_regression(family, solver, tol, fit_intercept, offset):
    """Test GLM regression with log link on a simple dataset."""
    coef = [0.2, -0.1]
    X = np.array([[1, 1, 1, 1, 1], [0, 1, 2, 3, 4]]).T
    y = np.exp(np.dot(X, coef) + (0 if offset is None else offset))
    glm = GeneralizedLinearRegressor(
        alpha=0,
        family=family,
        link="log",
        fit_intercept=fit_intercept,
        solver=solver,
        gradient_tol=tol,
    )
    if fit_intercept:
        X = X[:, 1:]
    res = glm.fit(X, y, offset=offset)
    if fit_intercept:
        fit_coef = np.concatenate([np.atleast_1d(res.intercept_), res.coef_])
    else:
        fit_coef = res.coef_
    assert_allclose(fit_coef, coef, rtol=8e-6)


@pytest.mark.filterwarnings("ignore:The line search algorithm")
@pytest.mark.filterwarnings("ignore:Line Search failed")
@pytest.mark.parametrize("n_samples, n_features", [(100, 10), (10, 100)])
@pytest.mark.parametrize("solver", GLM_SOLVERS)
@pytest.mark.parametrize("use_offset", [False, True])
def test_normal_ridge_comparison(n_samples, n_features, solver, use_offset):
    """Test ridge regression for Normal distributions.

    Case n_samples >> n_features

    Compare to test_ridge in test_ridge.py.
    """
    alpha = 1.0
    n_predict = 10
    X, y, _ = make_regression(
        n_samples=n_samples + n_predict,
        n_features=n_features,
        n_informative=n_features - 2,
        noise=0.5,
        coef=True,
        random_state=42,
    )
    y = y[0:n_samples]
    X, T = X[0:n_samples], X[n_samples:]
    if use_offset:
        np.random.seed(0)
        offset = np.random.randn(n_samples)
        y += offset
    else:
        offset = None

    if n_samples > n_features:
        ridge_params: Dict[str, Any] = {"solver": "svd"}
    else:
        ridge_params = {"solver": "sag", "max_iter": 10000, "tol": 1e-9}

    # GLM has 1/(2*n) * Loss + 1/2*L2, Ridge has Loss + L2
    ridge = Ridge(alpha=alpha * n_samples, random_state=42, **ridge_params)
    ridge.fit(X, y if offset is None else y - offset)

    glm = GeneralizedLinearRegressor(
        alpha=1.0,
        l1_ratio=0,
        family="normal",
        fit_intercept=True,
        max_iter=300,
        solver=solver,
        gradient_tol=1e-6,
        check_input=False,
        random_state=42,
    )
    glm.fit(X, y, offset=offset)
    assert glm.coef_.shape == (X.shape[1],)
    assert_allclose(glm.coef_, ridge.coef_, rtol=5e-5)
    assert_allclose(glm.intercept_, ridge.intercept_, rtol=1e-5)
    assert_allclose(glm.predict(T), ridge.predict(T), rtol=1e-4)


@pytest.mark.parametrize(
    "solver, tol",
    [("irls-ls", 1e-7), ("lbfgs", 1e-7), ("irls-cd", 1e-7), ("trust-constr", 1e-8)],
)
@pytest.mark.parametrize("scale_predictors", [True, False])
@pytest.mark.parametrize("use_sparse", [True, False])
def test_poisson_ridge(solver, tol, scale_predictors, use_sparse):
    """Test ridge regression with poisson family and LogLink.

    Compare to R's glmnet
    """
    # library("glmnet")
    # options(digits=10)
    # df <- data.frame(a=c(-2,-1,1,2), b=c(0,0,1,1), y=c(0,1,1,2))
    # x <- data.matrix(df[,c("a", "b")])
    # y <- df$y
    # fit <- glmnet(x=x, y=y, alpha=0, intercept=T, family="poisson",
    #               standardize=F, thresh=1e-10, nlambda=10000)
    # coef(fit, s=1)
    # (Intercept) -0.12889386979
    # a            0.29019207995
    # b            0.03741173122
    #
    # fit <- glmnet(x=x, y=y, alpha=0, intercept=T, family="poisson",
    #               standardize=T, thresh=1e-10, nlambda=10000)
    # coef(fit, s=1)
    # (Intercept) -0.21002571120839675
    # a            0.16472093,
    # b            0.27051971

    # Alternately, for running from Python:
    # from glmnet_python import glmnet
    # model = glmnet(x=X_dense, y=y, alpha=0, family="poisson",
    #               standardize=scale_predictors, thresh=1e-10, lambdau=np.array([1.0]))
    # true_intercept = model["a0"][0]
    # true_beta = model["beta"][:, 0]
    # print(true_intercept, true_beta)

    X_dense = np.array([[-2, -1, 1, 2], [0, 0, 1, 1]], dtype=np.float_).T
    if use_sparse:
        X = sparse.csc_matrix(X_dense)
    else:
        X = X_dense
    y = np.array([0, 1, 1, 2], dtype=np.float_)
    model_args = dict(
        alpha=1,
        l1_ratio=0,
        fit_intercept=True,
        family="poisson",
        link="log",
        gradient_tol=tol,
        solver=solver,
        max_iter=300,
        random_state=np.random.RandomState(42),
        copy_X=True,
        scale_predictors=scale_predictors,
    )
    glm = GeneralizedLinearRegressor(**model_args)
    glm2 = copy.deepcopy(glm)

    def check(G):
        G.fit(X, y)
        if scale_predictors:
            assert_allclose(G.intercept_, -0.21002571120839675, rtol=1e-5)
            assert_allclose(G.coef_, [0.16472093, 0.27051971], rtol=1e-5)
        else:
            assert_allclose(G.intercept_, -0.12889386979, rtol=1e-5)
            assert_allclose(G.coef_, [0.29019207995, 0.03741173122], rtol=1e-5)

    check(glm)

    # Test warm starting a re-fit model.
    glm.warm_start = True
    check(glm)
    assert glm.n_iter_ <= 1

    # Test warm starting with start_params.
    glm2.warm_start = True
    glm2.start_params = np.concatenate(([glm.intercept_], glm.coef_))
    check(glm2)
    assert glm2.n_iter_ <= 1


@pytest.mark.parametrize("scale_predictors", [True, False])
def test_poisson_ridge_bounded(scale_predictors):
    X = np.array([[-1, 1, 1, 2], [0, 0, 1, 1]], dtype=np.float_).T
    y = np.array([0, 1, 1, 2], dtype=np.float_)
    lb = np.array([-0.1, -0.1])
    ub = np.array([0.1, 0.1])

    # For comparison, this is the source of truth for the assert_allclose below.
    # from glmnet_python import glmnet
    # model = glmnet(x=X.copy(), y=y.copy(), alpha=0, family="poisson",
    #               standardize=scale_predictors, thresh=1e-10, lambdau=np.array([1.0]),
    #               cl = np.array([lb, ub])
    #               )
    # true_intercept = model["a0"][0]
    # true_beta = model["beta"][:, 0]
    # print(true_intercept, true_beta)

    glm = GeneralizedLinearRegressor(
        alpha=1,
        l1_ratio=0,
        fit_intercept=True,
        family="poisson",
        link="log",
        gradient_tol=1e-7,
        solver="irls-cd",
        max_iter=300,
        random_state=np.random.RandomState(42),
        copy_X=True,
        scale_predictors=scale_predictors,
        lower_bounds=lb,
        upper_bounds=ub,
    )
    glm.fit(X, y)

    # These correct values come from glmnet.
    assert_allclose(glm.intercept_, -0.13568186971946633, rtol=1e-5)
    assert_allclose(glm.coef_, [0.1, 0.1], rtol=1e-5)


@pytest.mark.parametrize("scale_predictors", [True, False])
def test_poisson_ridge_ineq_constrained(scale_predictors):
    X = np.array([[-1, 1, 1, 2], [0, 0, 1, 1]], dtype=np.float_).T
    y = np.array([0, 1, 1, 2], dtype=np.float_)
    A_ineq = np.array([[1, 0], [0, 1], [-1, 0], [0, -1]])
    b_ineq = 0.1 * np.ones(shape=(4))

    # For comparison, this is the source of truth for the assert_allclose below.
    # from glmnet_python import glmnet
    # model = glmnet(x=X.copy(), y=y.copy(), alpha=0, family="poisson",
    #               standardize=scale_predictors, thresh=1e-10, lambdau=np.array([1.0]),
    #               cl = np.array([lb, ub])
    #               )
    # true_intercept = model["a0"][0]
    # true_beta = model["beta"][:, 0]
    # print(true_intercept, true_beta)

    glm = GeneralizedLinearRegressor(
        alpha=1,
        l1_ratio=0,
        fit_intercept=True,
        family="poisson",
        link="log",
        gradient_tol=1e-12,  # 1e-8 not sufficient
        random_state=np.random.RandomState(42),
        copy_X=True,
        scale_predictors=scale_predictors,
        A_ineq=A_ineq,
        b_ineq=b_ineq,
    )
    glm.fit(X, y)

    # These correct values come from glmnet.
    assert_allclose(glm.intercept_, -0.13568186971946633, rtol=1e-5)
    assert_allclose(glm.coef_, [0.1, 0.1], rtol=1e-5)


def test_normal_enet():
    """Test elastic net regression with normal/gaussian family."""
    alpha, l1_ratio = 0.3, 0.7
    n_samples, n_features = 20, 2
    rng = np.random.RandomState(42)
    X = rng.randn(n_samples, n_features).copy(order="F")
    beta = rng.randn(n_features)
    y = 2 + np.dot(X, beta) + rng.randn(n_samples)

    # 1. test normal enet on dense data
    glm = GeneralizedLinearRegressor(
        alpha=alpha,
        l1_ratio=l1_ratio,
        family="normal",
        link="identity",
        fit_intercept=True,
        gradient_tol=1e-8,
        max_iter=100,
        selection="cyclic",
        solver="irls-cd",
        check_input=False,
    )
    glm.fit(X, y)

    enet = ElasticNet(
        alpha=alpha,
        l1_ratio=l1_ratio,
        fit_intercept=True,
        tol=1e-8,
        copy_X=True,
    )
    enet.fit(X, y)

    assert_allclose(glm.intercept_, enet.intercept_, rtol=2e-7)
    assert_allclose(glm.coef_, enet.coef_, rtol=5e-5)

    # 2. test normal enet on sparse data
    X = sparse.csc_matrix(X)
    glm.fit(X, y)
    assert_allclose(glm.intercept_, enet.intercept_, rtol=2e-7)
    assert_allclose(glm.coef_, enet.coef_, rtol=5e-5)


def test_poisson_enet():
    """Test elastic net regression with poisson family and LogLink.

    Compare to R's glmnet
    """
    # library("glmnet")
    # options(digits=10)
    # df <- data.frame(a=c(-2,-1,1,2), b=c(0,0,1,1), y=c(0,1,1,2))
    # x <- data.matrix(df[,c("a", "b")])
    # y <- df$y
    # fit <- glmnet(x=x, y=y, alpha=0.5, intercept=T, family="poisson",
    #               standardize=F, thresh=1e-10, nlambda=10000)
    # coef(fit, s=1)
    # (Intercept) -0.03550978409
    # a            0.16936423283
    # b            .
    glmnet_intercept = -0.03550978409
    glmnet_coef = [0.16936423283, 0.0]
    X = np.array([[-2, -1, 1, 2], [0, 0, 1, 1]]).T
    y = np.array([0, 1, 1, 2])
    rng = np.random.RandomState(42)
    glm = GeneralizedLinearRegressor(
        alpha=1,
        l1_ratio=0.5,
        family="poisson",
        link="log",
        solver="irls-cd",
        gradient_tol=1e-8,
        selection="random",
        random_state=rng,
    )
    glm.fit(X, y)
    assert_allclose(glm.intercept_, glmnet_intercept, rtol=2e-6)
    assert_allclose(glm.coef_, glmnet_coef, rtol=2e-7)

    # test results with general optimization procedure
    def obj(coef):
        pd = PoissonDistribution()
        link = LogLink()
        N = y.shape[0]
        mu = link.inverse(X @ coef[1:] + coef[0])
        alpha, l1_ratio = (1, 0.5)
        return (
            1.0 / (2.0 * N) * pd.deviance(y, mu)
            + 0.5 * alpha * (1 - l1_ratio) * (coef[1:] ** 2).sum()
            + alpha * l1_ratio * np.sum(np.abs(coef[1:]))
        )

    res = optimize.minimize(
        obj,
        [0, 0, 0],
        method="nelder-mead",
        tol=1e-10,
        options={"maxiter": 1000, "disp": False},
    )
    assert_allclose(glm.intercept_, res.x[0], rtol=5e-5)
    assert_allclose(glm.coef_, res.x[1:], rtol=1e-5, atol=1e-9)
    assert_allclose(
        obj(np.concatenate(([glm.intercept_], glm.coef_))), res.fun, rtol=1e-8
    )

    # same for start_params='zero' and selection='cyclic'
    # with reduced precision
    glm = GeneralizedLinearRegressor(
        alpha=1,
        l1_ratio=0.5,
        family="poisson",
        link="log",
        solver="irls-cd",
        gradient_tol=1e-5,
        selection="cyclic",
    )
    glm.fit(X, y)
    assert_allclose(glm.intercept_, glmnet_intercept, rtol=1e-4)
    assert_allclose(glm.coef_, glmnet_coef, rtol=1e-4)

    # check warm_start, therefore start with different alpha
    glm = GeneralizedLinearRegressor(
        alpha=0.005,
        l1_ratio=0.5,
        family="poisson",
        max_iter=300,
        link="log",
        solver="irls-cd",
        gradient_tol=1e-5,
        selection="cyclic",
    )
    glm.fit(X, y)
    # warm start with original alpha and use of sparse matrices
    glm.warm_start = True
    glm.alpha = 1
    X = sparse.csr_matrix(X)
    glm.fit(X, y)
    assert_allclose(glm.intercept_, glmnet_intercept, rtol=1e-4)
    assert_allclose(glm.coef_, glmnet_coef, rtol=1e-4)


@pytest.mark.parametrize("alpha", [0.01, 0.1, 1, 10])
def test_binomial_enet(alpha):
    """Test elastic net regression with binomial family and LogitLink.

    Compare to LogisticRegression.
    """
    l1_ratio = 0.5
    n_samples = 500
    rng = np.random.RandomState(42)
    X, y = make_classification(
        n_samples=n_samples,
        n_classes=2,
        n_features=6,
        n_informative=5,
        n_redundant=0,
        n_repeated=0,
        random_state=rng,
    )
    log = LogisticRegression(
        penalty="elasticnet",
        random_state=rng,
        fit_intercept=False,
        tol=1e-7,
        max_iter=1000,
        l1_ratio=l1_ratio,
        C=1.0 / (n_samples * alpha),
        solver="saga",
    )
    log.fit(X, y)

    glm = GeneralizedLinearRegressor(
        family=BinomialDistribution(),
        link=LogitLink(),
        fit_intercept=False,
        alpha=alpha,
        l1_ratio=l1_ratio,
        solver="irls-cd",
        selection="cyclic",
        gradient_tol=1e-7,
    )
    glm.fit(X, y)
    assert_allclose(log.intercept_[0], glm.intercept_, rtol=1e-6)
    assert_allclose(log.coef_[0, :], glm.coef_, rtol=5.1e-6)


@pytest.mark.parametrize(
    "params",
    [
        {"solver": "irls-ls"},
        {"solver": "lbfgs"},
        {"solver": "trust-constr"},
        {"solver": "irls-cd", "selection": "cyclic"},
        {"solver": "irls-cd", "selection": "random"},
    ],
    ids=lambda params: ", ".join(
        "{}={}".format(key, val) for key, val in params.items()
    ),
)
@pytest.mark.parametrize("use_offset", [False, True])
def test_solver_equivalence(params, use_offset, regression_data):
    X, y = regression_data
    if use_offset:
        np.random.seed(0)
        offset = np.random.random(len(y))
    else:
        offset = None
    est_ref = GeneralizedLinearRegressor(random_state=2)
    est_ref.fit(X, y, offset=offset)

    est_2 = GeneralizedLinearRegressor(**params)
    est_2.set_params(random_state=2)

    est_2.fit(X, y, offset=offset)

    assert_allclose(est_2.intercept_, est_ref.intercept_, rtol=1e-4)
    assert_allclose(est_2.coef_, est_ref.coef_, rtol=1e-4)
    assert_allclose(
        mean_absolute_error(est_2.predict(X), y),
        mean_absolute_error(est_ref.predict(X), y),
        rtol=1e-4,
    )


# TODO: different distributions
# Specify rtol since some are more accurate than others
@pytest.mark.parametrize(
    "params",
    [
        {"solver": "irls-ls", "rtol": 1e-6},
        {"solver": "lbfgs", "rtol": 2e-4},
        {"solver": "trust-constr", "rtol": 2e-4},
        {"solver": "irls-cd", "selection": "cyclic", "rtol": 2e-5},
        {"solver": "irls-cd", "selection": "random", "rtol": 6e-5},
    ],
    ids=lambda params: ", ".join(
        "{}={}".format(key, val) for key, val in params.items()
    ),
)
@pytest.mark.parametrize("use_offset", [False, True])
def test_solver_equivalence_cv(params, use_offset):
    n_alphas = 3
    n_samples = 100
    n_features = 10
    gradient_tol = 1e-5

    X, y = make_regression(n_samples=n_samples, n_features=n_features, random_state=2)
    if use_offset:
        np.random.seed(0)
        offset = np.random.random(len(y))
    else:
        offset = None

    est_ref = GeneralizedLinearRegressorCV(
        random_state=2,
        n_alphas=n_alphas,
        gradient_tol=gradient_tol,
        min_alpha_ratio=1e-3,
    )
    est_ref.fit(X, y, offset=offset)

    est_2 = (
        GeneralizedLinearRegressorCV(
            n_alphas=n_alphas,
            max_iter=1000,
            gradient_tol=gradient_tol,
            **{k: v for k, v in params.items() if k != "rtol"},
            min_alpha_ratio=1e-3,
        )
        .set_params(random_state=2)
        .fit(X, y, offset=offset)
    )

    def _assert_all_close(x, y):
        return assert_allclose(x, y, rtol=params["rtol"], atol=1e-7)

    _assert_all_close(est_2.alphas_, est_ref.alphas_)
    _assert_all_close(est_2.alpha_, est_ref.alpha_)
    _assert_all_close(est_2.l1_ratio_, est_ref.l1_ratio_)
    _assert_all_close(est_2.coef_path_, est_ref.coef_path_)
    _assert_all_close(est_2.deviance_path_, est_ref.deviance_path_)
    _assert_all_close(est_2.intercept_, est_ref.intercept_)
    _assert_all_close(est_2.coef_, est_ref.coef_)
    _assert_all_close(
        mean_absolute_error(est_2.predict(X), y),
        mean_absolute_error(est_ref.predict(X), y),
    )


@pytest.mark.parametrize("solver", GLM_SOLVERS)
def test_convergence_warning(solver, regression_data):
    X, y = regression_data

    est = GeneralizedLinearRegressor(
        solver=solver, random_state=2, max_iter=1, gradient_tol=1e-20
    )
    with pytest.warns(ConvergenceWarning):
        est.fit(X, y)


@pytest.mark.parametrize("use_sparse", [False, True])
@pytest.mark.parametrize("scale_predictors", [False, True])
def test_standardize(use_sparse, scale_predictors):
    def _arrays_share_data(arr1: np.ndarray, arr2: np.ndarray) -> bool:
        return arr1.__array_interface__["data"] == arr2.__array_interface__["data"]

    NR = 101
    NC = 10
    col_mults = np.arange(1, NC + 1)
    row_mults = np.linspace(0, 2, NR)
    M = row_mults[:, None] * col_mults[None, :]

    if use_sparse:
        M = tm.SparseMatrix(sparse.csc_matrix(M))
    else:
        M = tm.DenseMatrix(M)

    X, col_means, col_stds = M.standardize(np.ones(NR) / NR, True, scale_predictors)
    if use_sparse:
        assert _arrays_share_data(X.mat.data, M.data)
        assert _arrays_share_data(X.mat.indices, M.indices)
        assert _arrays_share_data(X.mat.indptr, M.indptr)
    else:
        # Check that the underlying data pointer is the same
        assert _arrays_share_data(X.mat, M)
    np.testing.assert_almost_equal(col_means, col_mults)

    # After standardization, all the columns will have the same values.
    # To check that, just convert to dense first.
    if use_sparse:
        Xdense = X.A
    else:
        Xdense = X
    for i in range(1, NC):
        if scale_predictors:
            if isinstance(Xdense, tm.StandardizedMatrix):
                one, two = Xdense.A[:, 0], Xdense.A[:, i]
            else:
                one, two = Xdense[:, 0], Xdense[:, i]
        else:
            if isinstance(Xdense, tm.StandardizedMatrix):
                one, two = (i + 1) * Xdense.A[:, 0], Xdense.A[:, i]
            else:
                one, two = (i + 1) * Xdense[:, 0], Xdense[:, i]
        np.testing.assert_almost_equal(one, two)

    if scale_predictors:
        # The sample variance of row_mults is 0.34. This is scaled up by the col_mults
        true_std = np.sqrt(0.34)
        np.testing.assert_almost_equal(col_stds, true_std * col_mults)

    intercept_standardized = 0.0
    coef_standardized = (
        np.ones_like(col_means) if col_stds is None else copy.copy(col_stds)
    )
    intercept, coef = _unstandardize(
        col_means,
        col_stds,
        intercept_standardized,
        coef_standardized,
    )
    np.testing.assert_almost_equal(intercept, -(NC + 1) * NC / 2)
    if scale_predictors:
        np.testing.assert_almost_equal(coef, 1.0)


@pytest.mark.parametrize("estimator, kwargs", estimators)
def test_check_estimator(estimator, kwargs):
    check_estimator(estimator(**kwargs))


@pytest.mark.parametrize(
    "estimator",
    [GeneralizedLinearRegressor, GeneralizedLinearRegressorCV],
)
def test_clonable(estimator):
    clone(estimator())


@pytest.mark.parametrize(
    "link, distribution, tol",
    [
        (IdentityLink(), NormalDistribution(), 1e-4),
        (LogLink(), PoissonDistribution(), 1e-4),
        (LogLink(), GammaDistribution(), 1e-4),
        (LogLink(), TweedieDistribution(1.5), 1e-4),
        (LogLink(), TweedieDistribution(4.5), 1e-4),
        (LogLink(), NormalDistribution(), 1e-4),
        (LogLink(), InverseGaussianDistribution(), 1e-4),
        (LogitLink(), BinomialDistribution(), 1e-2),
        (IdentityLink(), GeneralizedHyperbolicSecant(), 1e-1),
    ],
)
@pytest.mark.parametrize("offset", [None, np.array([0.3, -0.1, 0, 0.1]), 0.1])
def test_get_best_intercept(
    link: Link, distribution: ExponentialDispersionModel, tol: float, offset
):
    y = np.array([1, 1, 1, 2], dtype=np.float_)
    if isinstance(distribution, BinomialDistribution):
        y -= 1

    sample_weight = np.array([0.1, 0.2, 5, 1])
    best_intercept = guess_intercept(y, sample_weight, link, distribution, offset)
    assert np.isfinite(best_intercept)

    def _get_dev(intercept):
        eta = intercept if offset is None else offset + intercept
        mu = link.inverse(eta)
        assert np.isfinite(mu).all()
        return distribution.deviance(y, mu, sample_weight)

    obj = _get_dev(best_intercept)
    obj_low = _get_dev(best_intercept - tol)
    obj_high = _get_dev(best_intercept + tol)
    assert obj < obj_low
    assert obj < obj_high


@pytest.mark.parametrize("tol", [1e-2, 1e-4, 1e-6])
def test_step_size_tolerance(tol):
    X, y = make_regression(
        n_samples=100,
        n_features=5,
        noise=0.5,
        random_state=42,
    )
    y[y < 0] = 0

    def build_glm(step_size_tol):
        glm = GeneralizedLinearRegressor(
            alpha=1,
            l1_ratio=0.5,
            family="poisson",
            solver="irls-cd",
            gradient_tol=1e-10,
            step_size_tol=step_size_tol,
            selection="cyclic",
        )
        glm.fit(X, y)
        return glm

    baseline = build_glm(1e-10)
    glm = build_glm(tol)
    assert_allclose(baseline.intercept_, glm.intercept_, atol=tol)
    assert_allclose(baseline.coef_, glm.coef_, atol=tol)


def test_alpha_search(regression_data):
    X, y = regression_data
    mdl_no_path = GeneralizedLinearRegressor(
        alpha=0.001,
        l1_ratio=1,
        family="normal",
        link="identity",
        gradient_tol=1e-10,
    )
    mdl_no_path.fit(X=X, y=y)

    mdl_path = GeneralizedLinearRegressor(
        alpha_search=True,
        min_alpha=0.001,
        n_alphas=5,
        l1_ratio=1,
        family="normal",
        link="identity",
        gradient_tol=1e-10,
    )
    mdl_path.fit(X=X, y=y)

    assert_allclose(mdl_path.coef_, mdl_no_path.coef_)
    assert_allclose(mdl_path.intercept_, mdl_no_path.intercept_)


@pytest.mark.parametrize("alpha, alpha_index", [(0.5, 0), (0.75, 1), (None, 1)])
def test_predict_scalar(regression_data, alpha, alpha_index):

    X, y = regression_data
    offset = np.ones_like(y)

    estimator = GeneralizedLinearRegressor(alpha=[0.5, 0.75], alpha_search=True)
    estimator.fit(X, y)

    target = estimator.predict(X, alpha_index=alpha_index)

    candidate = estimator.predict(X, alpha=alpha, offset=offset)
    np.testing.assert_allclose(candidate, target + 1)


@pytest.mark.parametrize(
    "alpha, alpha_index",
    [([0.5, 0.75], [0, 1]), ([0.75, 0.5], [1, 0]), ([0.5, 0.5], [0, 0])],
)
def test_predict_list(regression_data, alpha, alpha_index):

    X, y = regression_data
    offset = np.ones_like(y)

    estimator = GeneralizedLinearRegressor(alpha=[0.5, 0.75], alpha_search=True)
    estimator.fit(X, y)

    target = np.stack(
        [
            estimator.predict(X, alpha_index=alpha_index[0]),
            estimator.predict(X, alpha_index=alpha_index[1]),
        ],
        axis=1,
    )

    candidate = estimator.predict(X, alpha=alpha, offset=offset)
    np.testing.assert_allclose(candidate, target + 1)

    candidate = estimator.predict(X, alpha_index=alpha_index, offset=offset)
    np.testing.assert_allclose(candidate, target + 1)


def test_predict_error(regression_data):

    X, y = regression_data

    estimator = GeneralizedLinearRegressor(alpha=0.5, alpha_search=False).fit(X, y)

    with pytest.raises(ValueError):
        estimator.predict(X, alpha=0.5)
    with pytest.raises(ValueError):
        estimator.predict(X, alpha=[0.5])
    with pytest.raises(AttributeError):
        estimator.predict(X, alpha_index=0)
    with pytest.raises(AttributeError):
        estimator.predict(X, alpha_index=[0])

    estimator.set_params(alpha=[0.5, 0.75], alpha_search=True).fit(X, y)

    with pytest.raises(IndexError):
        estimator.predict(X, y, alpha=0.25)
    with pytest.raises(IndexError):
        estimator.predict(X, y, alpha=[0.25, 0.5])
    with pytest.raises(IndexError):
        estimator.predict(X, y, alpha_index=2)
    with pytest.raises(IndexError):
        estimator.predict(X, y, alpha_index=[2, 0])
    with pytest.raises(ValueError):
        estimator.predict(X, y, alpha_index=0, alpha=0.5)


def test_very_large_initial_gradient():
    # this is a problem where 0 is a starting value that produces
    # a very large gradient initially
    np.random.seed(1234)
    y = np.exp(np.random.gamma(5, size=100))
    X = np.ones([len(y), 1])

    model_0 = GeneralizedLinearRegressor(
        link="log", family="gamma", fit_intercept=False, alpha=0, start_params=[0.0]
    ).fit(X, y)

    model_5 = GeneralizedLinearRegressor(
        link="log", family="gamma", fit_intercept=False, alpha=0, start_params=[5.0]
    ).fit(X, y)

    np.testing.assert_allclose(model_0.coef_, model_5.coef_, rtol=1e-5)


def test_fit_has_no_side_effects():
    y = np.array([0, 1, 2])
    w = np.array([0.5, 0.5, 0.5])
    X = np.array([[1, 1, 1]]).reshape(-1, 1)
    win = w.copy()
    yin = y.copy()
    Xin = X.copy()
    GeneralizedLinearRegressor(family="poisson").fit(Xin, yin, sample_weight=win)
    np.testing.assert_almost_equal(Xin, X)
    np.testing.assert_almost_equal(yin, y)
    np.testing.assert_almost_equal(win, w)
    GeneralizedLinearRegressor(family="poisson").fit(Xin, yin, offset=win)
    np.testing.assert_almost_equal(win, w)
    lb = np.array([-1.2])
    ub = np.array([1.2])
    lbin = lb.copy()
    ubin = ub.copy()
    GeneralizedLinearRegressor(
        family="poisson", scale_predictors=True, lower_bounds=lbin, upper_bounds=ubin
    ).fit(Xin, yin)
    np.testing.assert_almost_equal(lbin, lb)
    np.testing.assert_almost_equal(ubin, ub)


def test_column_with_stddev_zero():
    np.random.seed(1234)
    y = np.random.choice([1, 2, 3, 4], size=1000)
    X = np.ones([len(y), 1])

    model = GeneralizedLinearRegressor(
        family="poisson", fit_intercept=False, scale_predictors=False
    ).fit(
        X, y
    )  # noqa: F841
    model = GeneralizedLinearRegressor(family="poisson").fit(X, y)  # noqa: F841


@pytest.mark.parametrize("scale_predictors", [True, False])
@pytest.mark.parametrize("fit_intercept", [True, False])
@pytest.mark.parametrize("P1", ["identity", np.array([2, 1, 0.5, 0.1, 0.01])])
def test_alpha_path(scale_predictors, fit_intercept, P1):
    """Test regularization path."""
    if scale_predictors and not fit_intercept:
        return
    np.random.seed(1234)
    y = np.random.choice([1, 2, 3, 4], size=100)
    X = np.random.randn(100, 5) * np.array([1, 5, 10, 25, 100])

    model = GeneralizedLinearRegressor(
        family="poisson",
        alpha_search=True,
        l1_ratio=1,
        n_alphas=10,
        scale_predictors=scale_predictors,
        fit_intercept=fit_intercept,
        P1=P1,
    )
    model.fit(X=X, y=y)

    # maximum alpha result in all zero coefficients
    np.testing.assert_almost_equal(model.coef_path_[0], 0)
    # next alpha gives at least one non-zero coefficient
    assert np.any(model.coef_path_[1] > 0)


def test_passing_noncontiguous_as_X():
    X = np.random.rand(100, 4)
    y = np.random.rand(100)

    baseline = GeneralizedLinearRegressor(family="normal", fit_intercept=False).fit(
        X[:, :2].copy(), y
    )
    np_view = GeneralizedLinearRegressor(family="normal", fit_intercept=False).fit(
        X[:, :2], y
    )
    pd_view = GeneralizedLinearRegressor(family="normal", fit_intercept=False).fit(
        pd.DataFrame(X).iloc[:, :2], y
    )
    np.testing.assert_almost_equal(baseline.coef_, np_view.coef_)
    np.testing.assert_almost_equal(baseline.coef_, pd_view.coef_)


@pytest.mark.parametrize(
    "X, feature_names",
    [
        (pd.DataFrame({"x1": np.arange(5), "x2": 2}), np.array(["x1", "x2"])),
        (pd.DataFrame({"x1": np.arange(5), "x2": 2}).to_numpy(), None),
        (
            pd.DataFrame({"x1": pd.Categorical(np.arange(5)), "x2": 2}),
            np.array(["x1__0", "x1__1", "x1__2", "x1__3", "x1__4", "x2"]),
        ),
        (
            pd.DataFrame(
                {
                    "x1": pd.Categorical(np.arange(5)),
                    "x2": pd.Categorical([2, 2, 2, 2, 2]),
                }
            ),
            np.array(["x1__0", "x1__1", "x1__2", "x1__3", "x1__4", "x2__2"]),
        ),
    ],
)
def test_feature_names(X, feature_names):
    model = GeneralizedLinearRegressor(family="poisson").fit(X, np.arange(5))
    np.testing.assert_array_equal(getattr(model, "feature_names_", None), feature_names)


@pytest.mark.parametrize(
    "X, dtypes",
    [
        (pd.DataFrame({"x1": np.arange(5)}, dtype="int64"), {"x1": np.int64}),
        (pd.DataFrame({"x1": np.arange(5)}).to_numpy(), None),
        (
            pd.DataFrame({"x1": pd.Categorical(np.arange(5))}),
            {"x1": pd.CategoricalDtype(list(range(5)), ordered=False)},
        ),
    ],
)
def test_feature_dtypes(X, dtypes):
    model = GeneralizedLinearRegressor(family="poisson").fit(X, np.arange(5))
    np.testing.assert_array_equal(getattr(model, "feature_dtypes_", None), dtypes)


@pytest.mark.parametrize(
    "k, n",
    [
        (5, 5),
        (10, 5),
        (100, 5),
        (500, 50),
        (500, 100),
        (500, 500),
    ],
)
def test_categorical_types(k, n):
    np.random.seed(12345)
    categories = np.arange(k)
    group = np.random.choice(categories, size=n)
    y = group / k + np.random.uniform(size=n)

    # use categorical types
    X_cat = pd.DataFrame({"group": pd.Categorical(group, categories=categories)})
    model_cat = GeneralizedLinearRegressor(family="poisson").fit(X_cat, y)
    pred_cat = model_cat.predict(X_cat)

    # use one-hot encoding
    X_oh = pd.get_dummies(X_cat)
    model_oh = GeneralizedLinearRegressor(family="poisson").fit(X_oh, y)
    pred_oh = model_oh.predict(X_oh)

    # check predictions
    np.testing.assert_allclose(pred_cat, pred_oh)
    np.testing.assert_allclose(model_cat.intercept_, model_oh.intercept_)
    np.testing.assert_allclose(model_cat.coef_, model_oh.coef_)

    # compare across models/data types
    pred_cat_oh = model_cat.predict(X_oh)
    pred_oh_cat = model_oh.predict(X_cat)
    np.testing.assert_allclose(pred_cat_oh, pred_oh_cat)


@pytest.mark.parametrize(
    "kwargs", [{"alpha_search": True, "alpha": [1, 0.5, 0.1, 0.01]}, {"alpha": 0.1}]
)
def test_alpha_parametrization(kwargs, regression_data):
    X, y = regression_data
    model = GeneralizedLinearRegressor(**kwargs)
    model.fit(X=X, y=y)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"alpha_search": True, "alpha": 0.1},
        {"alpha_search": False, "alpha": [1, 0.5, 0.1, 0.01]},
        {"alpha_search": True, "alpha": "abc"},
        {"alpha": "abc"},
        {"alpha": -0.1},
        {"alpha_search": True, "alpha": [1, 0.5, -0.1]},
    ],
)
def test_alpha_parametrization_fail(kwargs, regression_data):
    X, y = regression_data
    with pytest.raises((ValueError, TypeError)):
        model = GeneralizedLinearRegressor(**kwargs)
        model.fit(X=X, y=y)


def test_verbose(regression_data, capsys):
    X, y = regression_data
    mdl = GeneralizedLinearRegressor(verbose=1)
    mdl.fit(X=X, y=y)
    captured = capsys.readouterr()
    assert "Iteration" in captured.err


def test_std_errors(regression_data):
    X, y = regression_data
    mdl = GeneralizedLinearRegressor(alpha=0, family="normal")
    mdl.fit(X=X, y=y)

    mdl_sm = sm.OLS(endog=y, exog=sm.add_constant(X))

    # nonrobust
    # Here, statsmodels does not do a degree of freedom adjustment,
    # so we manually add it.
    ourse = mdl.std_errors(X=X, y=y, robust=False)
    corr = len(y) / (len(y) - X.shape[1] - 1)
    smse = mdl_sm.fit(cov_type="nonrobust").bse * np.sqrt(corr)
    np.testing.assert_allclose(ourse, smse, rtol=1e-8)

    # robust
    ourse = mdl.std_errors(X=X, y=y, robust=True)
    smse = mdl_sm.fit(cov_type="HC1").bse
    np.testing.assert_allclose(ourse, smse, rtol=1e-8)

    # clustered
    rng = np.random.default_rng(42)
    clu = rng.integers(5, size=len(y))
    ourse = mdl.std_errors(X=X, y=y, clusters=clu)
    smse = mdl_sm.fit(cov_type="cluster", cov_kwds={"groups": clu}).bse
    np.testing.assert_allclose(ourse, smse, rtol=1e-8)


def test_sparse_std_errors(regression_data):
    X, y = regression_data
    sp_X = sparse.csc_matrix(X)
    mdl = GeneralizedLinearRegressor(alpha=0, family="normal")
    mdl.fit(X=X, y=y)

    actual1 = mdl.std_errors(X=sp_X, y=y, robust=False)
    expected1 = mdl.std_errors(X=X, y=y, robust=False)
    np.testing.assert_allclose(actual1, expected1)
    actual2 = mdl.std_errors(X=sp_X, y=y, robust=True)
    expected2 = mdl.std_errors(X=X, y=y, robust=True)
    np.testing.assert_allclose(actual2, expected2)
    rng = np.random.default_rng(42)
    clu = rng.integers(5, size=len(y))
    actual3 = mdl.std_errors(X=X, y=y, clusters=clu)
    expected3 = mdl.std_errors(X=X, y=y, clusters=clu)
    np.testing.assert_allclose(actual3, expected3)


@pytest.mark.parametrize("as_data_frame", [False, True])
@pytest.mark.parametrize("weighted", [False, True])
def test_score_method(as_data_frame, weighted):

    regressor = GeneralizedLinearRegressor(
        alpha=0,
        family="normal",
        fit_intercept=False,
        gradient_tol=1e-8,
        check_input=False,
    )

    y = np.array([-1, -1, 0, 1, 2])

    if weighted:
        y, wgts = np.unique(y, return_counts=True)
    else:
        wgts = None

    if as_data_frame:
        x = pd.DataFrame({"x": np.ones(len(y))})
    else:
        x = np.ones((len(y), 1))

    score = regressor.fit(x, y, sample_weight=wgts).score(x, y, sample_weight=wgts)

    # use pytest because NumPy used to always reject comparisons against zero
    assert pytest.approx(score, 1e-8) == 0



import glob
import os
import shutil
import time
from functools import reduce
from typing import Callable, Dict, Optional, Tuple, Union

import click
import numpy as np
import pandas as pd
import tabmat as tm
from scipy import sparse as sps

from glum import GeneralizedLinearRegressor, TweedieDistribution
from glum._solvers import eta_mu_objective

benchmark_convergence_tolerance = 1e-4
cache_location = os.environ.get("GLM_BENCHMARKS_CACHE", None)


def runtime(f, iterations, *args, **kwargs):
    """
    Measure how long it tales to run function f.

    Parameters
    ----------
    f: function
    iterations
    args: Passed to f
    kwargs: Passed to f

    Returns
    -------
    Tuple: (Minimimum runtime across iterations, output of f)

    """
    rs = []
    for _ in range(iterations):
        start = time.time()
        out = f(*args, **kwargs)
        end = time.time()
        rs.append(end - start)
    return np.min(rs), out


def get_sklearn_family(distribution):
    """
    Translate statistical family to its equivalent in sklearn jargon.

    Parameters
    ----------
    distribution
    """
    family = distribution
    if family == "gaussian":
        family = "normal"
    elif "tweedie" in family:
        tweedie_p = float(family.split("-p=")[1])
        family = TweedieDistribution(tweedie_p)  # type: ignore
    return family


def get_obj_val(
    dat: Dict[str, Union[np.ndarray, sps.spmatrix, tm.MatrixBase, pd.DataFrame]],
    distribution: str,
    alpha: float,
    l1_ratio: float,
    intercept: float,
    coefs: np.ndarray,
) -> float:
    """
    Return objective value.

    Parameters
    ----------
    dat
    distribution
    alpha
    l1_ratio
    intercept
    coefs

    Returns
    -------
    float
        Objective value

    """
    model = GeneralizedLinearRegressor(
        alpha=alpha,
        l1_ratio=l1_ratio,
        family=get_sklearn_family(distribution),
    )
    model._set_up_for_fit(dat["y"])

    full_coefs = np.concatenate(([intercept], coefs))
    offset = dat.get("offset")
    if isinstance(dat["X"], tm.MatrixBase):
        X_dot_coef = dat["X"].matvec(coefs)
    else:
        X_dot_coef = dat["X"].dot(coefs)
    X_dot_coef += intercept
    if isinstance(X_dot_coef, pd.Series):
        X_dot_coef = X_dot_coef.values
    if offset is not None:
        X_dot_coef += offset

    zeros = np.zeros(dat["X"].shape[0])
    y = dat["y"].astype(coefs.dtype)
    weights = dat.get("weights", np.ones_like(y)).astype(coefs.dtype)
    weights /= weights.sum()
    P1 = l1_ratio * alpha * np.ones_like(coefs)
    P2 = (1 - l1_ratio) * alpha * np.ones_like(coefs)

    _, _, obj_val, _ = eta_mu_objective(
        model._family_instance,
        model._link_instance,
        X_dot_coef,
        1.0,
        full_coefs,
        zeros,
        y,
        weights,
        P1,
        P2,
        1,
    )
    return obj_val


def exposure_and_offset_to_weights(
    power: Union[float, int],
    y: np.ndarray,
    exposure: np.ndarray = None,
    sample_weight: np.ndarray = None,
    offset: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Adjust outcomes and weights for exposure and offsets.

    This works for any Tweedie distributions with log-link. This is equivalence can be
    verified by checking the first order condition of the Tweedie log-likelihood.

    Parameters
    ----------
    power : float
        Power parameter of the Tweedie distribution.
    y : array-like
        Array with outcomes.
    exposure : array-like, optional, default=None
        Array with exposure.
    offset : array-like, optional, default=None
        Array with additive offsets.
    sample_weight : array-like, optional, default=None
        Array with sampling weights.

    Returns
    -------
    np.array
        Array with adjusted outcomes.
    np.array
        Estimation weights.
    """
    y = np.asanyarray(y)
    sample_weight = None if sample_weight is None else np.asanyarray(sample_weight)

    if offset is not None:
        exposure = np.exp(np.asanyarray(offset))
    elif exposure is not None:
        exposure = np.asarray(exposure)
    else:
        raise ValueError("Need offset or exposure.")
    y = y / exposure
    sample_weight = (
        exposure ** (2 - power)
        if sample_weight is None
        else sample_weight * exposure ** (2 - power)
    )
    return y, sample_weight


class BenchmarkParams:
    """
    Store metadata about the problem we want to solve and how.

    Attributes reflect exactly what the user passed in, only modified by type
    conversions. Any additional processing should be downstream.
    """

    def __init__(
        self,
        problem_name: Optional[str] = None,
        library_name: Optional[str] = None,
        num_rows: Optional[int] = None,
        storage: Optional[str] = None,
        threads: Optional[int] = None,
        single_precision: Optional[bool] = None,
        regularization_strength: Optional[float] = None,
        cv: Optional[bool] = None,
        hessian_approx: Optional[float] = None,
        diagnostics_level: Optional[str] = None,
    ):

        self.problem_name = problem_name
        self.library_name = library_name
        self.num_rows = num_rows
        self.storage = storage
        self.threads = threads
        self.single_precision = single_precision
        self.regularization_strength = regularization_strength
        self.cv = cv
        self.hessian_approx = hessian_approx
        self.diagnostics_level = diagnostics_level

    param_names = [
        "problem_name",
        "library_name",
        "num_rows",
        "storage",
        "threads",
        "single_precision",
        "regularization_strength",
        "cv",
        "hessian_approx",
        "diagnostics_level",
    ]

    def update_params(self, **kwargs):
        """
        Update attributes to those found in kwargs.

        Parameters
        ----------
        kwargs: dict

        Returns
        -------
        self with updated attributes

        """
        for k, v in kwargs.items():
            assert k in self.param_names
            setattr(self, k, v)
        return self

    def get_result_fname(self) -> str:
        """
        Get file name to which results will be written.

        Returns
        -------
        str
            File name within benchmark result directory.

        """
        return "_".join(str(getattr(self, k)) for k in self.param_names)


defaults = dict(
    threads=os.environ.get("OMP_NUM_THREADS", os.cpu_count()),
    problem_name=None,
    library_name=None,
    num_rows=None,
    regularization_strength=None,
    storage="dense",
    cv=False,
    single_precision=False,
    hessian_approx=0.0,
    diagnostics_level="basic",
)


def benchmark_params_cli(func: Callable) -> Callable:
    """
    Decorate a function so that options given via click CLI get mapped into a \
    BenchmarkParams instance.

    Parameters
    ----------
    func

    Returns
    -------
    Callable:
        wrapped function that takes a BenchmarkParams instance as an argument.

    """

    @click.option(
        "--problem_name",
        type=str,
        help="Specify a comma-separated list of benchmark problems you want to run. "
        "Leaving this blank will default to running all problems.",
    )
    @click.option(
        "--library_name",
        help="Specify a comma-separated list of libaries to benchmark. Leaving this "
        "blank will default to running all problems.",
    )
    @click.option(
        "--num_rows",
        type=int,
        help="Pass an integer number of rows. This is useful for testing and "
        "development. The default is to use the full dataset.",
    )
    @click.option(
        "--storage",
        type=str,
        help="Specify the storage format. Currently supported: dense, sparse. Leaving "
        "this black will default to dense.",
    )
    @click.option(
        "--threads",
        type=int,
        help="Specify the number of threads. If not set, it will use OMP_NUM_THREADS. "
        "If that's not set either, it will default to os.cpu_count().",
    )
    @click.option("--cv", type=bool, help="Cross-validation")
    @click.option("--single_precision", type=bool, help="Whether to use 32-bit data")
    @click.option(
        "--regularization_strength",
        type=float,
        help="Regularization strength. Set to None to use the default value of the "
        "problem.",
    )
    @click.option(
        "--hessian_approx",
        type=float,
        help="Threshold for dropping rows in the IRLS approximate Hessian update.",
    )
    @click.option(
        "--diagnostics_level",
        type=str,
        help="Choose 'basic' for brief glum diagnostics or 'full' for more "
        "extensive diagnostics. Any other string will result in no diagnostic "
        "output at all.",
    )
    def wrapped_func(
        problem_name: Optional[str],
        library_name: Optional[str],
        num_rows: Optional[int],
        storage: Optional[str],
        threads: Optional[int],
        cv: Optional[bool],
        single_precision: Optional[bool],
        regularization_strength: Optional[float],
        hessian_approx: Optional[float],
        diagnostics_level: Optional[str],
        *args,
        **kwargs,
    ):
        params = BenchmarkParams(
            problem_name,
            library_name,
            num_rows,
            storage,
            threads,
            single_precision,
            regularization_strength,
            cv,
            hessian_approx,
            diagnostics_level,
        )
        return func(params, *args, **kwargs)

    return wrapped_func


@click.command()
@benchmark_params_cli
def _get_params(params: BenchmarkParams):
    _get_params.out = params  # type: ignore


def get_params_from_fname(fname: str) -> BenchmarkParams:
    """
    Map file name to a BenchmarkParams instance.

    Parameters
    ----------
    fname: file name

    Returns
    -------
    BenchmarkParams

    """
    cli_list = reduce(
        lambda x, y: x + y,
        [
            ["--" + elt[0], elt[1]]
            for elt in zip(BenchmarkParams.param_names, fname.strip(".pkl").split("_"))
            if elt[1] != "None"
        ],
    )
    _get_params(cli_list, standalone_mode=False)
    return _get_params.out  # type: ignore


def _get_size_of_cache_directory():
    return sum(
        os.path.getsize(x) for x in glob.glob(f"{cache_location}/**", recursive=True)
    )


def clear_cache(force=False):
    """Clear the cache directory if its size exceeds a threshold."""
    if cache_location is None:
        return

    cache_size_limit = float(
        os.environ.get("GLM_BENCHMARKS_CACHE_SIZE_LIMIT", 1024 ** 3)
    )

    if force or _get_size_of_cache_directory() > cache_size_limit:
        shutil.rmtree(cache_location)


def get_tweedie_p(distribution: str) -> Union[float, int]:
    """
    Extract the "p" parameter of the Tweedie distribution from the string name of the \
    distribution.

    Examples
    --------
    >>> get_tweedie_p("tweedie_p=1.5")
    1.5

    """
    tweedie = "tweedie" in distribution
    if tweedie:
        return float(distribution.split("=")[-1])
    if "poisson" == distribution:
        return 1
    if "gamma" == distribution:
        return 2
    if "gaussian" in distribution:
        return 0
    else:
        raise ValueError("Not a Tweedie distribution.")



import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:
    __version__ = "unknown"

if "profile" not in __builtins__:  # type: ignore
    __builtins__["profile"] = lambda x: x  # type: ignore



import argparse
import copy
import json
import warnings

import numpy as np
import pytest
import tabmat as mx
from git_root import git_root
from scipy import sparse

from glum import GeneralizedLinearRegressor, GeneralizedLinearRegressorCV
from glum._glm import TweedieDistribution
from glum_benchmarks.data import simulate_glm_data

distributions_to_test = ["normal", "poisson", "gamma", "tweedie_p=1.5", "binomial"]
custom_family_link = [("normal", "log")]

# Not the canonical link for tweedie and gamma, but what the actuaries use.
link_map = {
    "normal": "identity",
    "poisson": "log",
    "gamma": "log",
    "tweedie_p=1.5": "log",
    "binomial": "logit",
}


def _make_P2():
    rand = np.random.default_rng(1)
    a = rand.uniform(size=(30, 30)) - 0.5  # centered uniform distribution
    P2 = a.T @ a  # make sure P2 is positive semi-definite
    return P2


@pytest.fixture(scope="module")
def data_all():
    return {
        dist: simulate_glm_data(
            family=dist,
            link=link_map[dist],
            n_rows=5000,
            dense_features=10,
            sparse_features=0,
            categorical_features=2,
            categorical_levels=10,
            ohe_categorical=True,
            drop_first=False,
        )
        for dist in distributions_to_test
    }


@pytest.fixture(
    params=["categorical"],
    scope="module",
)
def data_all_storage(request):
    data = {}
    for dist in distributions_to_test:
        data_config = {
            "family": dist,
            "link": link_map[dist],
            "n_rows": 5000,
            "dense_features": 10,
            "categorical_features": 2,
            "categorical_levels": 10,
        }

        if request.param == "dense":
            data_dist = simulate_glm_data(**data_config, ohe_categorical=True)
            data_dist["X"] = mx.DenseMatrix(data_dist["X"])
        elif request.param == "scipy-sparse":
            data_dist = simulate_glm_data(**data_config, ohe_categorical=True)
            data_dist["X"] = sparse.csc_matrix(data_dist["X"])
        elif request.param == "mkl-sparse":
            data_dist = simulate_glm_data(**data_config, ohe_categorical=True)
            data_dist["X"] = mx.SparseMatrix(sparse.csc_matrix(data_dist["X"]))
        elif request.param == "split":
            data_dist = simulate_glm_data(**data_config, ohe_categorical=True)
            data_dist["X"] = mx.csc_to_split(
                sparse.csc_matrix(data_dist["X"]), threshold=0.1
            )
        elif request.param == "categorical":
            data_dist = simulate_glm_data(**data_config, ohe_categorical=False)
            dense_X = mx.DenseMatrix(np.ascontiguousarray(data_dist["X"].iloc[:, :10]))
            cat0 = mx.CategoricalMatrix(data_dist["X"]["cat0"])
            cat1 = mx.CategoricalMatrix(data_dist["X"]["cat1"])
            data_dist["X"] = mx.SplitMatrix([dense_X, cat0, cat1])

        data[dist] = data_dist

    return data


@pytest.fixture(scope="module")
def expected_all():
    with open(git_root("tests/glm/golden_master/simulation_gm.json"), "r") as fh:
        return json.load(fh)


gm_model_parameters = {
    "default": {},  # default params
    "half-regularization": {"alpha": 0.5},  # regularization (other than alpha = 1)
    "elastic-net": {"l1_ratio": 0.5},  # elastic-net
    "lasso": {"l1_ratio": 1},  # lasso
    "variable_p1": {
        "l1_ratio": 1,
        "P1": np.arange(30) / 10,
    },  # lasso with variable penalty
    "variable_p2": {
        "l1_ratio": 0,
        "P2": _make_P2(),
    },  # ridge with Tikhonov regularization
    "variable_p1_p2": {
        "l1_ratio": 0.5,
        "P1": np.arange(30) / 10,
        "P2": _make_P2(),
    },  # elastic net with P1 and P2 variable penalty
    "fit_intercept": {"fit_intercept": False},  # do not fit the intercept
    "bounds": {"lower_bounds": np.full(30, 0), "upper_bounds": np.full(30, 0.4)},
}


def assert_gm_allclose(model, expected, rtol=0, atol=1e-5):
    np.testing.assert_allclose(
        model.coef_, np.array(expected["coef_"]), rtol=rtol, atol=atol
    )
    np.testing.assert_allclose(
        model.intercept_, expected["intercept_"], rtol=rtol, atol=atol
    )


def fit_model(data, family, model_parameters, use_weights, use_offset, cv):
    if "tweedie" in family:
        p = float(family.split("=")[1])
        family = TweedieDistribution(power=p)
    if cv:
        model = GeneralizedLinearRegressorCV(
            family=family, gradient_tol=1e-6, **model_parameters
        )
    else:
        model = GeneralizedLinearRegressor(
            family=family, gradient_tol=1e-6, **model_parameters
        )

    fit_params = {
        "X": data["X"],
        "y": data["y"],
    }
    if use_weights:
        fit_params.update({"sample_weight": data["sample_weight"]})
    if use_offset:
        fit_params.update({"offset": data["offset"]})

    model.fit(**fit_params)
    return model


@pytest.mark.parametrize(
    "distribution",
    distributions_to_test,
    ids=distributions_to_test,
)
@pytest.mark.parametrize(
    ["run_name", "model_parameters"],
    gm_model_parameters.items(),
    ids=gm_model_parameters.keys(),
)
@pytest.mark.parametrize("use_weights", [True, False], ids=["weights", "no_weights"])
@pytest.mark.parametrize("use_offset", [True, False], ids=["offset", "no_offset"])
def test_gm_features(
    distribution,
    model_parameters,
    run_name,
    use_weights,
    use_offset,
    data_all,
    expected_all,
):
    data = data_all[distribution]
    model = fit_model(
        data=data,
        family=distribution,
        model_parameters=model_parameters,
        use_weights=use_weights,
        use_offset=use_offset,
        cv=False,
    )

    if use_weights:
        run_name = f"{run_name}_weights"
    if use_offset:
        run_name = f"{run_name}_offset"
    expected = expected_all[distribution][run_name]

    assert_gm_allclose(model, expected)


@pytest.mark.parametrize(
    "distribution",
    distributions_to_test,
    ids=distributions_to_test,
)
def test_gm_storage(distribution, data_all_storage, expected_all):
    data = data_all_storage[distribution]
    model = fit_model(
        data=data,
        family=distribution,
        model_parameters={},
        use_weights=False,
        use_offset=False,
        cv=False,
    )

    run_name = "default"
    expected = expected_all[distribution][run_name]

    assert_gm_allclose(model, expected)


@pytest.mark.parametrize("family_link", custom_family_link)
@pytest.mark.parametrize("use_weights", [True, False], ids=["weights", "no_weights"])
@pytest.mark.parametrize("use_offset", [True, False], ids=["offset", "no_offset"])
def test_gm_custom_link(family_link, use_weights, use_offset, data_all, expected_all):
    """Currently only testing log-linear model."""
    distribution, link = family_link
    data = data_all[distribution]
    model_parameters = {"link": link}
    model = fit_model(
        data=data,
        family=distribution,
        model_parameters=model_parameters,
        use_weights=use_weights,
        use_offset=use_offset,
        cv=False,
    )

    run_name = f"custom-{distribution}-{link}"
    if use_weights:
        run_name = f"{run_name}_weights"
    if use_offset:
        run_name = f"{run_name}_offset"
    expected = expected_all[distribution][run_name]

    assert_gm_allclose(model, expected)


@pytest.mark.parametrize(
    "distribution",
    distributions_to_test,
    ids=distributions_to_test,
)
@pytest.mark.parametrize("use_weights", [True, False], ids=["weights", "no_weights"])
@pytest.mark.parametrize("use_offset", [True, False], ids=["offset", "no_offset"])
def test_gm_approx_hessian(
    distribution, use_weights, use_offset, data_all, expected_all
):
    data = data_all[distribution]
    model_parameters = {
        "hessian_approx": 0.1,
    }
    model = fit_model(
        data=data,
        family=distribution,
        model_parameters=model_parameters,
        use_weights=use_weights,
        use_offset=use_offset,
        cv=False,
    )

    run_name = "default"
    if use_weights:
        run_name = f"{run_name}_weights"
    if use_offset:
        run_name = f"{run_name}_offset"

    expected = expected_all[distribution][run_name]
    assert_gm_allclose(model, expected)


@pytest.mark.parametrize(
    "distribution",
    distributions_to_test,
    ids=distributions_to_test,
)
def test_gm_cv(distribution, data_all, expected_all):
    data = data_all[distribution]
    model_parameters = {
        "alphas": [0.1, 0.05, 0.01],
        "l1_ratio": [0.2, 0.5, 0.9],
        "cv": 3,
    }
    model = fit_model(
        data=data,
        family=distribution,
        model_parameters=model_parameters,
        use_weights=False,
        use_offset=False,
        cv=True,
    )

    run_name = f"CV-{distribution}"

    expected = expected_all[distribution][run_name]
    assert_gm_allclose(model, expected)


@pytest.mark.parametrize(
    "dist_power",
    [("poisson", 1), ("gamma", 2), ("tweedie_p=1.5", 1.5)],
)
def test_weights_offset_equivalence(dist_power, data_all):
    distribution, power = dist_power

    weights_data = data_all[distribution]
    offset_data = copy.copy(weights_data)

    exposure = np.exp(weights_data["offset"])
    weights_data["sample_weight"] = exposure ** (2 - power)
    weights_data["y"] = weights_data["y"] / exposure

    weights_parameters = {"alpha": 0.1 / weights_data["sample_weight"].mean()}
    model_weights = fit_model(
        data=weights_data,
        family=distribution,
        model_parameters=weights_parameters,
        use_weights=True,
        use_offset=False,
        cv=False,
    )

    offset_parameters = {"alpha": 0.1}
    model_offset = fit_model(
        data=offset_data,
        family=distribution,
        model_parameters=offset_parameters,
        use_weights=False,
        use_offset=True,
        cv=False,
    )

    np.testing.assert_allclose(
        model_weights.coef_, model_offset.coef_, rtol=0, atol=1e-5
    )
    np.testing.assert_allclose(
        model_weights.intercept_, model_offset.intercept_, rtol=0, atol=1e-5
    )


def run_and_store_golden_master(
    distribution,
    model_parameters,
    run_name,
    use_weights,
    use_offset,
    cv,
    data,
    gm_dict,
    overwrite=False,
):
    if use_weights:
        run_name = f"{run_name}_weights"
    if use_offset:
        run_name = f"{run_name}_offset"

    if distribution not in gm_dict.keys():
        gm_dict[distribution] = {}

    if run_name in gm_dict[distribution].keys():
        if overwrite:
            warnings.warn("Overwriting existing result")
        else:
            warnings.warn("Result exists and cannot overwrite. Skipping")
            return gm_dict

    model = fit_model(
        data=data,
        family=distribution,
        model_parameters=model_parameters,
        use_weights=use_weights,
        use_offset=use_offset,
        cv=cv,
    )

    gm_dict[distribution][run_name] = dict(
        coef_=model.coef_.tolist(),
        intercept_=model.intercept_,
        n_iter_=model.n_iter_,  # not used right now but could in the future
    )
    return gm_dict


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--overwrite", action="store_true", help="overwrite existing golden master"
    )
    args = parser.parse_args()

    try:
        with open(git_root("tests/glm/golden_master/simulation_gm.json"), "r") as fh:
            gm_dict = json.load(fh)
    except FileNotFoundError:
        gm_dict = {}

    for dist in distributions_to_test:
        data = simulate_glm_data(family=dist, link=link_map[dist])
        for mdl_param in gm_model_parameters.items():
            for use_weights in [True, False]:
                for use_offset in [True, False]:
                    gm_dict = run_and_store_golden_master(
                        distribution=dist,
                        model_parameters=mdl_param[1],
                        run_name=mdl_param[0],
                        use_weights=use_weights,
                        use_offset=use_offset,
                        cv=False,
                        data=data,
                        gm_dict=gm_dict,
                        overwrite=args.overwrite,
                    )

    for dist in distributions_to_test:
        data = simulate_glm_data(family=dist, link=link_map[dist])
        gm_dict = run_and_store_golden_master(
            distribution=dist,
            model_parameters={
                "alphas": [0.1, 0.05, 0.01],
                "l1_ratio": [0.2, 0.5, 0.9],
                "cv": 3,
            },
            run_name=f"CV-{dist}",
            use_weights=use_weights,
            use_offset=use_offset,
            cv=True,
            data=data,
            gm_dict=gm_dict,
            overwrite=args.overwrite,
        )

    for dist, link in custom_family_link:
        data = simulate_glm_data(family=dist, link=link_map[dist])
        for use_weights in [True, False]:
            for use_offset in [True, False]:
                gm_dict = run_and_store_golden_master(
                    distribution=dist,
                    model_parameters={"link": link},
                    run_name=f"custom-{dist}-{link}",
                    use_weights=use_weights,
                    use_offset=use_offset,
                    cv=False,
                    data=data,
                    gm_dict=gm_dict,
                    overwrite=args.overwrite,
                )

    with open(git_root("tests/glm/golden_master/simulation_gm.json"), "w") as fh:
        json.dump(gm_dict, fh, indent=2)



import time
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from scipy import sparse as sps

from glum import GeneralizedLinearRegressor, GeneralizedLinearRegressorCV

from .util import benchmark_convergence_tolerance, get_sklearn_family, runtime

random_seed = 110


def _build_and_fit(model_args, fit_args, cv: bool):
    if cv:
        return GeneralizedLinearRegressorCV(**model_args).fit(**fit_args)
    return GeneralizedLinearRegressor(**model_args).fit(**fit_args)


def glum_bench(
    dat: Dict[str, Union[np.ndarray, sps.spmatrix]],
    distribution: str,
    alpha: float,
    l1_ratio: float,
    iterations: int,
    cv: bool,
    diagnostics_level: str = "basic",
    reg_multiplier: Optional[float] = None,
    hessian_approx: float = 0.0,
    **kwargs,
):
    """
    Run the glum.GeneralizedLinearRegressor benchmark.

    Parameters
    ----------
    dat
    distribution
    alpha
    l1_ratio
    iterations
    cv
    diagnostics_level
    reg_multiplier
    hessian_approx
    kwargs

    Returns
    -------
    dict

    """
    result = {}

    X = dat["X"]
    fit_args = dict(X=X, y=dat["y"])
    if "sample_weight" in dat.keys():
        fit_args.update({"sample_weight": dat["sample_weight"]})
    if "offset" in dat.keys():
        fit_args.update({"offset": dat["offset"]})

    model_args = dict(
        family=get_sklearn_family(distribution),
        l1_ratio=l1_ratio,
        max_iter=1000,
        random_state=random_seed,
        copy_X=False,
        selection="cyclic",
        gradient_tol=1 if cv else benchmark_convergence_tolerance,
        step_size_tol=0.01 * benchmark_convergence_tolerance,
        force_all_finite=False,
        hessian_approx=hessian_approx,
        verbose=False,
    )

    if not cv:
        model_args["alpha"] = (
            alpha if reg_multiplier is None else alpha * reg_multiplier
        )

    result["runtime"], m = runtime(_build_and_fit, iterations, model_args, fit_args, cv)
    if not cv:
        # Just check that predict works here... This doesn't take very long.
        m.predict(**{k: v for k, v in fit_args.items() if k != "y"})

    result["intercept"] = m.intercept_
    result["coef"] = m.coef_
    result["n_iter"] = m.n_iter_
    if cv:
        alphas: np.ndarray = m.alphas_
        result["n_alphas"] = len(alphas)
        result["max_alpha"] = alphas.max()
        result["min_alpha"] = alphas.min()
        result["best_alpha"] = m.alpha_

    with pd.option_context(
        "display.expand_frame_repr", False, "max_columns", None, "max_rows", None
    ):
        if diagnostics_level == "basic":
            m.report_diagnostics()
        elif diagnostics_level == "full":
            m.report_diagnostics(full_report=True)
    return result


def _compute_path(niters, model_args, fit_args):
    step_model_args = model_args.copy()
    step_model_args["max_iter"] = 1
    step_model_args["warm_start"] = True
    m = GeneralizedLinearRegressor(**step_model_args)
    path = []
    for _ in range(niters):
        start = time.time()
        m.fit(**fit_args)
        print(time.time() - start)
        path.append(m.coef_.copy())
    return path



import os
import pickle
from typing import Any, Dict, List, Optional

import click
import numpy as np
import pandas as pd

from glum_benchmarks.problems import get_all_problems
from glum_benchmarks.util import (
    BenchmarkParams,
    benchmark_params_cli,
    clear_cache,
    get_params_from_fname,
)


def _get_comma_sep_names(xs: str) -> List[str]:
    return [x.strip() for x in xs.split(",")]


@click.command()
@click.option(
    "--output_dir",
    default="benchmark_output",
    help="The directory where we load benchmarking output.",
)
@click.option(
    "--export",
    default=None,
    type=str,
    help="File name or path to export the results to CSV or Pickle.",
)
@click.option(
    "--cols", default=None, type=str, help="Which output analysis columns to display?"
)
@benchmark_params_cli
def cli_analyze(
    params: BenchmarkParams, output_dir: str, export: Optional[str], cols: str
):
    """
    Describe runtime, objective function values, and other statistics on the \
    already-run problems specified by the command line options.

    Parameters
    ----------
    params: BenchmarkParams
    output_dir: str
    export: string or None
    cols: str
    """
    clear_cache()
    display_precision = 4
    np.set_printoptions(precision=display_precision, suppress=True)
    pd.set_option("precision", display_precision)

    file_names = _identify_parameter_fnames(output_dir, params)

    raw_results = {
        fname: _load_benchmark_results(output_dir, fname) for fname in file_names
    }
    formatted_results = [
        _extract_dict_results_to_pd_series(name, res)
        for name, res in raw_results.items()
        if len(res) > 0
    ]

    if not formatted_results:
        return

    res_df = pd.DataFrame.from_records(formatted_results)
    res_df["offset"] = res_df["problem_name"].apply(lambda x: "offset" in x)
    res_df["problem_name"] = [
        "weights".join(x.split("offset")) for x in res_df["problem_name"]
    ]
    problem_id_cols = ["problem_name", "num_rows", "regularization_strength", "offset"]
    res_df = res_df.set_index(problem_id_cols).sort_values("library_name").sort_index()
    if params.cv:
        for col in ["max_alpha", "min_alpha"]:
            res_df[col] = res_df[col].astype(float)

    res_df["rel_obj_val"] = (
        res_df[["obj_val"]] - res_df.groupby(level=[0, 1, 2, 3])[["obj_val"]].min()
    )

    with pd.option_context(
        "display.expand_frame_repr", False, "max_columns", None, "max_rows", None
    ):
        if cols is not None:
            cols_to_show = _get_comma_sep_names(cols)
        else:
            cols_to_show = [
                "library_name",
                "storage",
                "threads",
                "single_precision",
                "n_iter",
                "runtime",
            ]
            if res_df["cv"].any():
                cols_to_show += ["n_alphas", "max_alpha", "min_alpha", "best_alpha"]
            else:
                cols_to_show += [
                    "intercept",
                    "num_nonzero_coef",
                    "obj_val",
                    "rel_obj_val",
                ]
        if "library_name" not in cols_to_show:
            cols_to_show.insert(0, "library_name")
        print(res_df[cols_to_show])

    if export:
        if export.endswith(".pkl"):
            res_df.to_pickle(export)
        else:
            res_df.to_csv(export)

    return res_df


def _extract_dict_results_to_pd_series(
    fname: str,
    results: Dict[str, Any],
) -> Dict:
    assert "coef" in results.keys()
    params = get_params_from_fname(fname)
    assert params.problem_name is not None

    coefs = results["coef"]
    runtime_per_iter = results["runtime"] / results["n_iter"]
    l1_norm = np.sum(np.abs(coefs))
    l2_norm = np.sum(coefs ** 2)
    num_nonzero_coef = np.sum(np.abs(coefs) > 1e-8)

    # weights and offsets are solving the same problem, but the objective is set up to
    # deal with weights, so load the data for the weights problem rather than the
    # offset problem
    if "housing" not in params.problem_name:
        prob_name_weights = "weights".join(params.problem_name.split("offset"))
    else:
        prob_name_weights = params.problem_name
    problem = get_all_problems()[prob_name_weights]

    formatted: Dict[str, Any] = params.__dict__
    items_to_use_from_results = ["n_iter", "runtime", "intercept"]
    if params.cv:
        items_to_use_from_results += [
            "n_alphas",
            "max_alpha",
            "min_alpha",
            "best_alpha",
        ]
    formatted.update(
        {k: v for k, v in results.items() if k in items_to_use_from_results}
    )

    formatted.update(
        {
            "num_rows": results["num_rows"],
            "regularization_strength": (
                problem.regularization_strength
                if params.regularization_strength is None
                else params.regularization_strength
            ),
            "runtime per iter": runtime_per_iter,
            "l1": l1_norm,
            "l2": l2_norm,
            "num_nonzero_coef": num_nonzero_coef,
            "obj_val": results["obj_val"],
            "offset": "offset" in params.problem_name,
        }
    )

    return formatted


def _identify_parameter_fnames(
    root_dir: str, constraint_params: BenchmarkParams
) -> List[str]:
    def _satisfies_constraint(params: BenchmarkParams, k: str) -> bool:
        constraint = getattr(constraint_params, k)
        param = getattr(params, k)
        return (
            constraint is None
            or param == constraint
            # e.g. this_file_params['library_name'] is 'sklearn-fork'
            # and constraint_params.library_name is 'sklearn-fork,h2o'
            or (isinstance(constraint, str) and param in constraint.split(","))
        )

    results_to_keep = []
    for fname in os.listdir(root_dir):
        this_file_params = get_params_from_fname(fname)

        keep_this_problem = {
            k: _satisfies_constraint(this_file_params, k)
            for k in constraint_params.param_names
        }

        if all(keep_this_problem.values()):
            results_to_keep.append(fname)
    return results_to_keep


def _load_benchmark_results(output_dir: str, fname: str):
    results_path = os.path.join(output_dir, fname)
    with open(results_path, "rb") as f:
        return pickle.load(f)


if __name__ == "__main__":
    cli_analyze()



import os
from typing import Optional, Tuple

import numpy as np
import pandas as pd
from git_root import git_root
from sklearn.datasets import fetch_openml


def create_housing_raw_data() -> None:
    """Do some basic processing on the data that we will later transform into our \
    benchmark data sets."""
    # Load the dataset from sklearn
    house_data = fetch_openml(name="house_sales", version=3, as_frame=True)

    # Use only select features
    df_house = house_data.data[
        [
            "bedrooms",
            "bathrooms",
            "sqft_living",
            "floors",
            "waterfront",
            "view",
            "condition",
            "grade",
            "yr_built",
            "yr_renovated",
        ]
    ].copy()

    # Targets
    df_house["price"] = house_data.target
    df_house["above_median_price"] = (
        df_house["price"] < df_house["price"].median()
    ).astype("int")

    # Save
    out_path = git_root("data/housing.parquet")
    os.makedirs(os.path.dirname(out_path), exist_ok=True)
    df_house.to_parquet(git_root("data/housing.parquet"))


def add_noise(df: pd.DataFrame, noise: float) -> pd.DataFrame:
    """Add noise by swapping out data points."""
    np.random.seed(43212)
    for col in df.columns:
        if col in ["price", "above_median_price"]:
            continue
        swap = np.random.uniform(size=len(df)) < noise
        shuffle = np.random.choice(df[col], size=len(df))
        df.loc[swap, col] = shuffle[swap]

    return df


def compute_y_exposure(df, distribution):
    """Compute y and exposure depending on distribution.

    (Exposure/weights for housing data always all 1).
    """
    if distribution in ["gamma", "gaussian"]:
        y = df["price"].values
    elif distribution == "binomial":
        y = df["above_median_price"].values
    else:
        raise ValueError(
            f"distribution for housing problems must be one of"
            f"['gamma', 'gaussian', 'binomial'] not {distribution}."
        )

    return y, np.ones_like(y)


def _read_housing_data(
    num_rows: Optional[int], noise: Optional[float], distribution: str
) -> pd.DataFrame:
    path = git_root("data/housing.parquet")
    if not os.path.exists(path):
        create_housing_raw_data()
    df = pd.read_parquet(path)

    if num_rows is not None:
        # if we're oversampling, set default value for noise to 0.05
        # can be turned off by setting noise to zero
        if noise is None and num_rows > len(df):
            noise = 0.05
        df = df.sample(n=num_rows, replace=True, random_state=12345)

    if noise is not None:
        df = add_noise(df, noise=noise)
    return df


def generate_housing_dataset(
    num_rows=None, noise=None, distribution="poisson"
) -> Tuple[pd.DataFrame, np.ndarray, np.ndarray]:
    """Generate the openml house_sales housing dataset."""
    df = _read_housing_data(num_rows, noise, distribution)

    y, exposure = compute_y_exposure(df, distribution)

    return df, y, exposure


if __name__ == "__main__":
    create_housing_raw_data()



import openml


def download_and_transform():
    """Download data from openml and apply basic transformations."""
    dataset = openml.datasets.get_dataset(42092)
    df, y, _, _ = dataset.get_data(
        dataset_format="dataframe", target=dataset.default_target_attribute
    )
    df["price"] = y

    df = df[(df["price"] < 1.5e6) & (df["price"] > 1e5)]

    drop_cols = [
        "date",
        "sqft_lot",
        "grade",
        "sqft_above",
        "yr_renovated",
        "lat",
        "long",
        "sqft_living15",
        "sqft_lot15",
    ]
    df = df.drop(columns=drop_cols)
    return df



import time
from typing import Any, Callable, Dict, Tuple

import numpy as np
import pandas as pd
import tabmat as tm
from scipy import sparse as sps
from tabmat.ext.dense import dense_sandwich
from tabmat.ext.sparse import sparse_sandwich

from .problems import (
    generate_narrow_insurance_dataset,
    generate_wide_insurance_dataset,
    load_data,
)


def _load(which: str, n_rows: int) -> Tuple[Any, np.ndarray]:
    if which == "narrow":
        x = sps.csc_matrix(load_data(generate_narrow_insurance_dataset, n_rows)["X"])
    else:
        x = sps.csc_matrix(load_data(generate_wide_insurance_dataset, n_rows)["X"])
    np.random.seed(0)
    d = np.random.uniform(0, 1, n_rows)
    return x, d


def _naive_sandwich(x, d):
    return (x.T @ (x.multiply(d[:, None]))).toarray()


def _fast_sandwich(X, d):
    return sparse_sandwich(X, X.XT, d)


def _split_sandwich(X, threshold):
    Xsplit = tm.from_split(X, threshold)
    return lambda _, d: Xsplit.sandwich(d)


def _dense_sandwich(X, d):
    return dense_sandwich(X.X_dense, d)


def _run_one_problem_all_methods(x, d, include_naive, dtype) -> pd.DataFrame:
    x = x.astype(dtype)
    d = d.astype(dtype)
    x.XT = x.T.tocsc()
    x.X_dense = np.asfortranarray(x.toarray())
    funcs: Dict[str, Callable[[Any, np.ndarray], Any]] = {
        "sparse_sandwich": _fast_sandwich,
        "dense_sandwich": _dense_sandwich,
    }
    funcs["split_sandwich_0.05"] = _split_sandwich(x, 0.05)
    funcs["split_sandwich_0.1"] = _split_sandwich(x, 0.1)
    # for threshold in [0.0, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3,0.4,0.5,0.7,0.9,1.0]:
    # for threshold in [0.01, 0.02, 0.03, 0.04, 0.05]:
    #     funcs[f"split_sandwich_{threshold}"] = split_sandwich(x, threshold)

    if include_naive:
        funcs["naive"] = _naive_sandwich

    info: Dict[str, Any] = {}
    for name, func in funcs.items():
        ts = []
        for _ in range(7):
            start = time.perf_counter()
            res = func(x, d)
            ts.append(time.perf_counter() - start)
        elapsed = np.min(ts)

        info[name] = {}
        info[name]["res"] = res
        info[name]["time"] = elapsed

    if include_naive:
        naive_result = info["naive"]["res"]
        for k in info:
            np.testing.assert_allclose(naive_result, info[k]["res"], 4)

    return pd.DataFrame(
        {"method": list(info.keys()), "time": [elt["time"] for elt in info.values()]}
    )


def main() -> None:
    """Run sparse sandwich benchmarks."""
    # "killed" with 1e7 and 4 * 1e6
    row_counts = [
        int(1e4),
        # int(1e5),
        # int(3e5),
        int(1e6),
        # int(2e6),
    ]  # , int(2e6), int(4e6), int(10e6)]
    benchmarks = []

    x, d = _load("narrow", row_counts[-1])

    for i, n_rows in enumerate(row_counts):
        for dtype in [np.float32, np.float64]:
            benchmarks.append(
                _run_one_problem_all_methods(
                    x[:n_rows, :].copy(), d[:n_rows].copy(), i == 0, dtype
                )
            )
            benchmarks[-1]["dtype"] = str(dtype)
            benchmarks[-1]["n_rows"] = n_rows

    result_df = pd.concat(benchmarks).sort_values(["n_rows", "method"])
    print(result_df.set_index(["n_rows", "method"]))


if __name__ == "__main__":
    main()



import os
import warnings
from typing import Dict, Optional, Union

import h2o
import numpy as np
import pandas as pd
from h2o.estimators.glm import H2OGeneralizedLinearEstimator
from scipy import sparse as sps

from .util import benchmark_convergence_tolerance, runtime


def _build_and_fit(model_args, train_args):
    glm = H2OGeneralizedLinearEstimator(**model_args)
    glm.train(**train_args)
    return glm


def _hstack_sparse_or_dense(to_stack):
    if sps.isspmatrix(to_stack[0]):
        return sps.hstack(to_stack)
    else:
        return np.hstack(to_stack)


def h2o_bench(
    dat: Dict[str, Union[np.ndarray, sps.spmatrix]],
    distribution: str,
    alpha: float,
    l1_ratio: float,
    iterations: int,
    cv: bool,
    reg_multiplier: Optional[float] = None,
    **kwargs,
):
    """
    Run a benchmark problem using h2o's glm.

    Parameters
    ----------
    dat
    distribution
    alpha
    l1_ratio
    iterations
    cv
    reg_multiplier
    kwargs

    Returns
    -------
    dict of data about this run
    """
    result: Dict = {}

    if not isinstance(dat["X"], (np.ndarray, sps.spmatrix, pd.DataFrame)):
        warnings.warn(
            "h2o requires data as scipy.sparse matrix, pandas dataframe, or numpy "
            "array. Skipping."
        )
        return result

    h2o.init(nthreads=int(os.environ.get("OMP_NUM_THREADS", os.cpu_count())))  # type: ignore

    train_mat = _hstack_sparse_or_dense((dat["X"], dat["y"][:, np.newaxis]))

    use_weights = "sample_weight" in dat.keys()
    if use_weights:
        train_mat = _hstack_sparse_or_dense(
            (train_mat, dat["sample_weight"][:, np.newaxis])
        )
    if "offset" in dat.keys():
        train_mat = _hstack_sparse_or_dense((train_mat, dat["offset"][:, np.newaxis]))

    train_h2o = h2o.H2OFrame(train_mat)

    tweedie = "tweedie" in distribution

    model_args = dict(
        model_id="glm",
        # not sure if this is right
        family="tweedie" if tweedie else distribution,
        alpha=l1_ratio,
        lambda_=alpha if reg_multiplier is None else alpha * reg_multiplier,
        standardize=False,
        solver="IRLSM",
        objective_epsilon=benchmark_convergence_tolerance,
        beta_epsilon=benchmark_convergence_tolerance,
        gradient_epsilon=benchmark_convergence_tolerance,
    )
    if cv:
        model_args["lambda_search"] = True
        model_args["nfolds"] = 5

    if tweedie:
        p = float(distribution.split("=")[-1])
        model_args["tweedie_variance_power"] = p
        model_args["tweedie_link_power"] = 1 if p == 0 else 0
    if "gamma" in distribution:
        model_args["link"] = "Log"

    if use_weights:
        train_args = dict(
            x=train_h2o.col_names[:-2],
            y=train_h2o.col_names[-2],
            training_frame=train_h2o,
            weights_column=train_h2o.col_names[-1],
        )
    elif "offset" in dat.keys():
        train_args = dict(
            x=train_h2o.col_names[:-2],
            y=train_h2o.col_names[-2],
            training_frame=train_h2o,
            offset_column=train_h2o.col_names[-1],
        )
    else:
        train_args = dict(
            x=train_h2o.col_names[:-1],
            y=train_h2o.col_names[-1],
            training_frame=train_h2o,
        )

    result["runtime"], m = runtime(_build_and_fit, iterations, model_args, train_args)
    # un-standardize
    standardized_intercept = m.coef()["Intercept"]

    standardized_coefs = np.array(
        [
            # h2o automatically removes zero-variance columns; impute to 1
            m.coef().get(f"C{i + 1}", 0)
            for i in range(
                train_mat.shape[1] - (2 if use_weights or "offset" in dat.keys() else 1)
            )
        ]
    )
    if cv:
        result["best_alpha"] = m._model_json["output"]["lambda_best"]
        result["n_alphas"] = m.parms["nlambdas"]["actual_value"]

    result["intercept"] = standardized_intercept
    result["coef"] = standardized_coefs

    result["n_iter"] = m.score_history().iloc[-1]["iteration" if cv else "iterations"]
    return result



"""
Helpers for loading map data.

Some functions adapted from:
https://towardsdatascience.com/
mapping-with-matplotlib-pandas-geopandas-and-basemap-in-python-d11b57ab5dac
"""

import numpy as np
import pandas as pd
import shapefile as shp
import shapely


def read_shapefile(shp_path):
    """Create data frame from shapefile.

    Parameters
    ----------
    shp_path : string
        Path to ``.shp`` file for map.

    Returns
    -------
    df : pandas.DataFrame
        Data frame with postcodes and corresponding geometric features.
    """
    sf = shp.Reader(shp_path)
    fields = [x[0] for x in sf.fields][1:]
    records = [list(i) for i in sf.records()]
    shps = [_get_shapely_shape(s) for s in sf.shapes()]
    df = pd.DataFrame(columns=fields, data=records)
    df = df.assign(geometry=shps)
    return df


def _get_shapely_shape(s):
    if len(s.parts) == 1:
        return shapely.geometry.Polygon(s.points, [])
    return shapely.geometry.multipolygon.MultiPolygon(
        [[ps] for ps in np.split(s.points, s.parts[1:])], context_type="geojson"
    )


def _get_polygon_coords(x):
    if type(x) is shapely.geometry.multipolygon.MultiPolygon:
        return [list(x[j].exterior.coords) for j in range(len(x.geoms))]
    else:
        return [list(x.exterior.coords)]


def create_kings_county_map_df(df, df_shapefile):
    """Create data frame with sales and geometric information.

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame with housing sales information per postcode.
    df_shapefile: pandas.DataFrame
        Data frame with map geometry information per postcode.

    Returns
    -------
    df : pandas.DataFrame
        Data frame with sales and geometric information. Includes coordinates
        in array form and points for region centroids (for labeling).
    """
    df_shapefile["ZIP"] = df_shapefile["ZIP"].astype(str)

    # include certain regions that have no data to prevent "holes" in map
    df_shapefile = df_shapefile[
        df_shapefile["ZIP"].isin(
            list(df.zipcode.unique()) + ["98051", "98158", "98057"]
        )
    ]
    # dissolve boundary between shared zip codes
    df_shapefile = (
        df_shapefile.groupby(by="ZIP")
        .agg({"geometry": shapely.ops.unary_union, "ZIPCODE": "first"})
        .reset_index()
    )
    df_shapefile["centroids"] = df_shapefile["geometry"].apply(
        lambda x: (x.centroid.coords[0][0] - 0.015, x.centroid.coords[0][1])
    )
    df_shapefile["coords"] = df_shapefile["geometry"].apply(
        lambda x: _get_polygon_coords(x)
    )
    return df_shapefile.merge(
        df.groupby(["zipcode"])["price"].mean(),
        left_on="ZIP",
        right_on="zipcode",
        how="outer",
    )


def _calc_color(data, vmin, vmax):
    color_sq = [
        "#F5EEF8",
        "#EBDEF0",
        "#D7BDE2",
        "#C39BD3",
        "#AF7AC5",
        "#9B59B6",
        "#884EA0",
        "#76448A",
        "#633974",
    ]
    new_data = data.copy()
    if (vmin is not None) or (vmax is not None):
        new_data = np.clip(new_data, vmin, vmax)
    new_data, _ = pd.cut(new_data, 9, retbins=True, labels=list(range(9)))
    return [color_sq[val] if not pd.isna(val) else None for val in new_data]


def plot_heatmap(df, label_col, data_col, ax, vmin=None, vmax=None):
    """Plot geographic heatmap on provided axes.

    Parameters
    ----------
    df : pandas.DataFrame
        Data frame with geometric and sales information (or any other
        attributes to determine heat map intensity).
    label_col: str
        The column of the data frame to use as labels on the heatmap.
    data_col: string
        The column of the data frame with which to determine heatmap intensity.
    ax: matplotlib.axes._subplots.AxesSubplot
        Matplotlib axes to plot on.
    vmin, vmax: float, optional (default=None))
        The range covered by the heatmap. With default (``None``),
        complete range of data covered.
    """
    df["color_ton"] = _calc_color(df[data_col], vmin, vmax)
    for point_sets in df.coords:
        for points in point_sets:
            x = [i[0] for i in points[:]]
            y = [i[1] for i in points[:]]
            ax.plot(x, y, "k")
    for _, row in df.iterrows():
        for dim in range(len(row.coords)):
            x_lon = np.zeros((len(row.coords[dim]), 1))
            y_lat = np.zeros((len(row.coords[dim]), 1))
            for ip in range(len(row.coords[dim])):
                x_lon[ip] = row.coords[dim][ip][0]
                y_lat[ip] = row.coords[dim][ip][1]
            if row.color_ton is None:
                ax.fill(x_lon, y_lat, "#FFFFFF", hatch="///", edgecolor="#747474")
            else:
                ax.fill(x_lon, y_lat, row.color_ton)
        x0 = row["centroids"][0]
        y0 = row["centroids"][1]
        ax.text(x0, y0, row[label_col], fontsize=10)



import numpy as np
import pandas as pd


def compute_zscore(df, window=100, min_periods=None, outcome="sales", id="store"):
    """Compute rolling zscore."""
    mu = (
        df[[id, outcome]]
        .groupby(id)[outcome]
        .rolling(window, min_periods=min_periods)
        .mean()
        .groupby(level=0)
        .shift()
        .droplevel(0)
    )
    sd = (
        df[[id, outcome]]
        .groupby(id)[outcome]
        .rolling(window, min_periods=min_periods)
        .std(ddof=0)
        .groupby(level=0)
        .shift()
        .droplevel(0)
    )
    return ((df[outcome] - mu) / sd).replace([np.inf, -np.inf], np.nan)


def compute_age_quantile(df, cuts=5):
    """Compute age quantiles."""
    cumcount = df.groupby("store")["date"].transform("cumcount")
    potential_age = (df["date"] - df["date"].min()).dt.days
    return pd.qcut(
        cumcount.where(potential_age > cumcount), cuts, duplicates="drop"
    ).cat.codes


def compute_competition_open(df):
    """Create a competitor open feature."""
    # competition open?
    missing_mask = (
        df["competition_open_since_year"].isna()
        | df["competition_open_since_month"].isna()
    )
    competition_open = (
        pd.to_datetime(
            df["competition_open_since_year"].astype("Int64").astype("str")
            + "-"
            + df["competition_open_since_month"]
            .astype("Int64")
            .astype(str)
            .str.zfill(2)
            + "-01",
            errors="coerce",
        )
        .le(df["date"])
        .astype(int)
    )
    return competition_open.where(~missing_mask).astype("category")


def compute_lags(df, outcome="sales", suffix="", offset=1, **kwargs):
    """Compute moving averages."""
    df[f"ewm_{outcome}_lag{suffix}"] = (
        df.assign(value=lambda x: x[outcome].where(x["open"].eq(1)))[["store", "value"]]
        .groupby("store")
        .ewm(**kwargs)
        .mean()["value"]
        .groupby(level=0)
        .shift(offset)
        .droplevel(0)
    ).to_numpy()

    df[f"ewm_{outcome}_lag{suffix}_day_of_week"] = np.nan
    for day_of_week in range(1, 8):
        roll = (
            df.assign(
                value=lambda x: x[outcome].where(
                    x["day_of_week"].eq(day_of_week) & x["open"].eq(1)
                )
            )[["store", "value"]]
            .groupby("store")
            .ewm(**kwargs)
            .mean()["value"]
            .groupby(level=0)
            .shift(offset)
            .droplevel(0)
        ).to_numpy()
        mask = df["day_of_week"].eq(day_of_week).to_numpy()
        df.loc[mask, f"ewm_{outcome}_lag{suffix}_day_of_week"] = roll[mask]

    df[f"ewm_{outcome}_lag{suffix}_month"] = np.nan
    for month in range(1, 13):
        roll = (
            df.assign(
                value=lambda x: x[outcome].where(x["month"].eq(month) & x["open"].eq(1))
            )[["store", "value"]]
            .groupby("store")
            .ewm(**kwargs)
            .mean()["value"]
            .groupby(level=0)
            .shift(offset)
            .droplevel(0)
        ).to_numpy()
        mask = df["month"].eq(month).to_numpy()
        df.loc[mask, f"ewm_{outcome}_lag{suffix}_month"] = roll[mask]

    return df


def compute_open_lead(df, shift=-1):
    """Compute leading indicator for open."""
    open_lead = df.groupby("store")["open"].shift(shift)
    return open_lead.combine_first(df["day_of_week"].ne(6).astype("double"))


def compute_open_lag(df, shift=1):
    """Compute lagged indicator for open."""
    open_lag = df.groupby("store")["open"].shift(shift)
    return open_lag.combine_first(df["day_of_week"].ne(1).astype("double"))


def compute_school_holiday_lead(df, shift=-1):
    """Compute leading indicator for school_holiday."""
    school_holiday_lead = df.groupby("store")["school_holiday"].shift(shift)
    return school_holiday_lead.fillna(0).astype("double")


def compute_school_holiday_lag(df, shift=1):
    """Compute lagged indicator for school_holiday."""
    school_holiday_lag = df.groupby("store")["school_holiday"].shift(shift)
    return school_holiday_lag.fillna(0).astype("double")


def compute_state_holiday_lead(df, shift=-1):
    """Compute leading indicator for state_holiday."""
    state_holiday_lead = df.groupby("store")["state_holiday"].shift(shift)
    return state_holiday_lead.fillna("0").astype(object)


def compute_state_holiday_lag(df, shift=1):
    """Compute lagged indicator for state_holiday."""
    state_holiday_lag = df.groupby("store")["state_holiday"].shift(shift)
    return state_holiday_lag.fillna("0").astype(object)


def compute_promo_lead(df, shift=-1):
    """Compute leading indicator for promo."""
    promo_lead = df.groupby("store")["promo"].shift(shift)
    return promo_lead.fillna(0)


def compute_promo_lag(df, shift=1):
    """Compute lagged indicator for promo."""
    promo_lag = df.groupby("store")["promo"].shift(shift)
    return promo_lag.fillna(0)


def compute_store_state_holiday(df):
    """Compute store state holiday interaction."""
    return df["store"].astype(str) + "_" + df["state_holiday"].ne("0").astype(str)


def compute_store_school_holiday(df):
    """Compute store school vacation interaction effect."""
    return df["store"].astype(str) + "_" + df["school_holiday"].astype(str)


def compute_store_year(df):
    """Compute store year fixed effect."""
    return df["store"].astype(str) + "_" + df["year"].astype(str)


def compute_store_month(df):
    """Compute store month fixed effect."""
    return df["store"].astype(str) + "_" + df["month"].astype(str)


def compute_store_day_of_week(df):
    """Compute store day of week fixed effect."""
    return df["store"].astype(str) + "_" + df["day_of_week"].astype(str)


def compute_lagged_ma(df, lag=48, window=48, min_periods=48):
    """Compute lagged moving average."""
    lagged_ma = (
        df[["store", "sales"]]
        .groupby("store")["sales"]
        .rolling(window, min_periods=min_periods)
        .sum()
        .groupby(level=0)
        .shift(lag)
        .droplevel(0)
    )
    lagged_open = (
        df[["store", "open"]]
        .groupby("store")["open"]
        .rolling(window, min_periods=min_periods)
        .sum()
        .groupby(level=0)
        .shift(lag)
        .droplevel(0)
    )
    return lagged_ma / lagged_open


def compute_lagged_ewma(
    df, lag=48, com=None, span=None, halflife=None, alpha=None, min_periods=50
):
    """Compute lagged exponentially weighted moving average."""
    lagged_ma = (
        df[["store", "sales"]]
        .assign(sales=lambda x: x["sales"].replace(0, np.nan))
        .groupby("store")["sales"]
        .ewm(
            com=com,
            span=span,
            halflife=halflife,
            alpha=alpha,
            min_periods=min_periods,
            ignore_na=True,
        )
        .mean()
        .groupby(level=0)
        .shift(lag)
        .droplevel(0)
    )
    return lagged_ma


def apply_all_transformations(df):
    """Apply all feature transformations."""
    df["age_quantile"] = compute_age_quantile(df, 5)
    df["competition_open"] = compute_competition_open(df)
    df["count"] = df.groupby("store")[["date"]].transform("cumcount")
    df["open_lag_1"] = compute_open_lag(df)
    df["open_lag_2"] = compute_open_lag(df, 2)
    df["open_lag_3"] = compute_open_lag(df, 3)
    df["open_lead_1"] = compute_open_lead(df)
    df["open_lead_2"] = compute_open_lead(df, -2)
    df["open_lead_3"] = compute_open_lead(df, -3)
    df["promo_lag_1"] = compute_promo_lag(df)
    df["promo_lag_2"] = compute_promo_lag(df, 2)
    df["promo_lag_3"] = compute_promo_lag(df, 3)
    df["promo_lead_1"] = compute_promo_lead(df)
    df["promo_lead_2"] = compute_promo_lead(df, -2)
    df["promo_lead_3"] = compute_promo_lead(df, -3)
    df["school_holiday_lag_1"] = compute_school_holiday_lag(df)
    df["school_holiday_lag_2"] = compute_school_holiday_lag(df, 2)
    df["school_holiday_lag_3"] = compute_school_holiday_lag(df, 3)
    df["school_holiday_lead_1"] = compute_school_holiday_lead(df)
    df["school_holiday_lead_2"] = compute_school_holiday_lead(df, -2)
    df["school_holiday_lead_3"] = compute_school_holiday_lead(df, -3)
    df["state_holiday_lag_1"] = compute_state_holiday_lag(df)
    df["state_holiday_lag_2"] = compute_state_holiday_lag(df, 2)
    df["state_holiday_lag_3"] = compute_state_holiday_lag(df, 3)
    df["state_holiday_lead_1"] = compute_state_holiday_lead(df)
    df["state_holiday_lead_2"] = compute_state_holiday_lead(df, -2)
    df["state_holiday_lead_3"] = compute_state_holiday_lead(df, -3)
    df["store_day_of_week"] = compute_store_day_of_week(df)
    df["store_month"] = compute_store_month(df)
    df["store_school_holiday"] = compute_store_school_holiday(df)
    df["store_state_holiday"] = compute_store_state_holiday(df)
    df["store_year"] = compute_store_year(df)
    df["zscore"] = compute_zscore(df, window=150)

    return df



import warnings
from typing import Any, Dict, Optional, Union

import numpy as np
import pandas as pd
from scipy import sparse as sps
from sklearn.linear_model import LogisticRegression

from .util import benchmark_convergence_tolerance, runtime


def _build_and_fit(model_args, train_args):
    return LogisticRegression(**model_args).fit(**train_args)


def liblinear_bench(
    dat: Dict[str, Union[sps.spmatrix, np.ndarray]],
    distribution: str,
    alpha: float,
    l1_ratio: float,
    iterations: int,
    cv: bool,
    reg_multiplier: Optional[float] = None,
    **kwargs,
) -> Dict[str, Any]:
    """
    Run the benchmark for sklearn.linear_model.LogisticRegression.

    Parameters
    ----------
    dat
    distribution
    alpha
    l1_ratio
    iterations
    cv
    reg_multiplier
    kwargs

    Returns
    -------
    dict

    """
    result: Dict = {}

    X = dat["X"]
    if not isinstance(X, (np.ndarray, sps.spmatrix, pd.DataFrame)):
        warnings.warn(
            "liblinear requires data as scipy.sparse matrix, pandas dataframe, or numpy "
            "array. Skipping."
        )
        return result

    if distribution != "binomial":
        warnings.warn("liblinear only supports binomial")
        return result

    if l1_ratio == 1 and alpha > 0:
        pen = "l1"
    elif l1_ratio == 0 and alpha > 0:
        pen = "l2"
    else:
        warnings.warn(
            "liblinear only supports lasso and ridge regression with positive alpha"
        )
        return result

    if "offset" in dat.keys():
        warnings.warn("liblinear does not support offsets")
        return result

    if cv:
        warnings.warn("liblinear does not yet support CV")
        return result

    model_args = dict(
        penalty=pen,
        tol=benchmark_convergence_tolerance,
        C=1 / (X.shape[0] * alpha)
        if reg_multiplier is None
        else 1 / (X.shape[0] * alpha * reg_multiplier),
        # Note that when an intercept is fitted, it is subject to regularization, unlike
        # other solvers. intercept_scaling helps combat this by inflating the intercept
        # column, though too low of a value leaves too much regularization and too high
        # of a value results in poor matrix properties.
        # See https://scikit-learn.org/stable/modules/generated/
        # sklearn.linear_model.LogisticRegression.html
        intercept_scaling=1e3,
        solver="liblinear",
    )

    fit_args = dict(
        X=X,
        y=dat["y"].astype(np.int64).copy(),
        sample_weight=dat.get("sample_weight"),
    )

    result["runtime"], m = runtime(_build_and_fit, iterations, model_args, fit_args)
    result["intercept"] = m.intercept_[0]
    result["coef"] = np.squeeze(m.coef_)
    result["n_iter"] = m.n_iter_[0]

    return result



import os
from functools import partial
from typing import Callable, Dict, List, Optional, Tuple, Union

import attr
import numpy as np
import pandas as pd
import tabmat as tm
from dask_ml.preprocessing import DummyEncoder
from git_root import git_root
from joblib import Memory
from scipy.sparse import csc_matrix

from .data import (
    generate_housing_dataset,
    generate_intermediate_insurance_dataset,
    generate_narrow_insurance_dataset,
    generate_real_insurance_dataset,
    generate_wide_insurance_dataset,
)
from .util import cache_location, exposure_and_offset_to_weights, get_tweedie_p

joblib_memory = Memory(cache_location, verbose=0)


@attr.s
class Problem:
    """Store metadata about which problem we should run."""

    data_loader = attr.ib(type=Callable)
    distribution = attr.ib(type=str)
    regularization_strength = attr.ib(type=float)
    l1_ratio = attr.ib(type=float)


@joblib_memory.cache
def load_data(
    loader_func: Callable[
        [Optional[int], Optional[float], Optional[str]],
        Tuple[pd.DataFrame, np.ndarray, np.ndarray],
    ],
    num_rows: int = None,
    storage: str = "dense",
    single_precision: bool = False,
    noise: float = None,
    distribution: str = "poisson",
    data_setup: str = "weights",
) -> Dict[str, np.ndarray]:
    """
    Load the data.

    A note about weights and exposures: Due to the way we have set up this problem, by
    rescaling the target variable, it is appropriate to pass what is modeled as an
    'exposure' as a weight. Everywhere else, exposures will be referred to as weights.
    """
    # TODO: add a weights_and_offset option
    # Step 1) Load the data.
    if data_setup not in ["weights", "offset", "no-weights"]:
        raise NotImplementedError
    X_in, y, exposure = loader_func(num_rows, noise, distribution)

    # Step 2) Convert to needed precision level.
    if single_precision:
        X_in = X_in.astype(np.float32)
        y = y.astype(np.float32)
        if exposure is not None:
            exposure = exposure.astype(np.float32)

    # Step 3) One hot encode columns if we are not using CategoricalMatrix
    def transform_col(i: int, dtype) -> Union[pd.DataFrame, tm.CategoricalMatrix]:
        if dtype.name == "category":
            if storage == "cat":
                return tm.CategoricalMatrix(X_in.iloc[:, i])
            return DummyEncoder().fit_transform(X_in.iloc[:, [i]])
        return X_in.iloc[:, [i]]

    mat_parts = [transform_col(i, dtype) for i, dtype in enumerate(X_in.dtypes)]
    # TODO: add a threshold for the number of categories needed to make a categorical
    #  matrix

    # Step 4) Convert the matrix to the appopriate storage type.
    if storage == "auto":
        dtype = np.float32 if single_precision else np.float64
        X = tm.from_pandas(X_in, dtype, sparse_threshold=0.1, cat_threshold=3)
    elif storage == "cat":
        cat_indices_in_expanded_arr: List[np.ndarray] = []
        dense_indices_in_expanded_arr: List[int] = []
        i = 0
        for elt in mat_parts:
            assert elt.ndim == 2
            if isinstance(elt, tm.CategoricalMatrix):
                ncol = elt.shape[1]
                cat_indices_in_expanded_arr.append(np.arange(i, i + ncol))
                i += ncol
            else:
                dense_indices_in_expanded_arr.append(i)
                i += 1

        non_cat_part = tm.DenseMatrix(
            np.hstack(
                [
                    elt.values
                    for elt in mat_parts
                    if not isinstance(elt, tm.CategoricalMatrix)
                ]
            )
        )
        X = tm.SplitMatrix(
            matrices=[non_cat_part]
            + [elt for elt in mat_parts if isinstance(elt, tm.CategoricalMatrix)],
            indices=[np.array(dense_indices_in_expanded_arr)]
            + cat_indices_in_expanded_arr,
        )
    elif storage == "sparse":
        X = csc_matrix(pd.concat(mat_parts, axis=1))
    elif storage.startswith("split"):
        threshold = float(storage.split("split")[1])
        X = tm.from_csc(csc_matrix(pd.concat(mat_parts, axis=1)), threshold)
    else:  # Fall back to using a dense matrix.
        X = pd.concat(mat_parts, axis=1)

    # Step 5) Handle weights or offsets if needed.
    if data_setup == "weights":
        # The exposure correction doesn't make sense for these distributions since
        # they don't use a log link (plus binomial isn't in the tweedie family),
        # but this is what we were doing before.
        if distribution in ["gaussian", "binomial"]:
            return dict(X=X, y=y, sample_weight=exposure)
        # when poisson, should be y=y, sample_weight=exposure
        # instead have y = y / exposure, weight = exposure
        y, sample_weight = exposure_and_offset_to_weights(
            get_tweedie_p(distribution), y, exposure
        )
        return dict(X=X, y=y * exposure, sample_weight=sample_weight)

    if data_setup == "offset":
        log_exposure = np.log(exposure)
        assert np.all(np.isfinite(log_exposure))
        # y has already been divided by exposure loader_func, so undo it here
        return dict(X=X, y=y * exposure, offset=log_exposure)
    # data_setup = "no_weights"
    return dict(X=X, y=y)


def get_all_problems() -> Dict[str, Problem]:
    """
    Return the name of all possible problems.

    Returns
    -------
    Dict mapping problem names to Problem instances.

    """
    regularization_strength = 0.001

    housing_distributions = ["gaussian", "gamma", "binomial"]
    housing_load_funcs = {
        "intermediate-housing": generate_housing_dataset,
    }

    insurance_distributions = [
        "gaussian",
        "poisson",
        "gamma",
        "tweedie-p=1.5",
        "binomial",
    ]
    insurance_load_funcs = {
        "intermediate-insurance": generate_intermediate_insurance_dataset,
        "narrow-insurance": generate_narrow_insurance_dataset,
        "wide-insurance": generate_wide_insurance_dataset,
    }
    if os.path.isfile(git_root("data", "X.parquet")):
        insurance_load_funcs["real-insurance"] = generate_real_insurance_dataset

    problems = {}
    for penalty_str, l1_ratio in [("l2", 0.0), ("net", 0.5), ("lasso", 1.0)]:
        # Add housing problems
        for distribution in housing_distributions:
            suffix = penalty_str + "-" + distribution
            dist = distribution
            for problem_name, load_fn in housing_load_funcs.items():
                for data_setup in ["no-weights", "offset"]:
                    problems["-".join((problem_name, data_setup, suffix))] = Problem(
                        data_loader=partial(
                            load_data, load_fn, distribution=dist, data_setup=data_setup
                        ),
                        distribution=distribution,
                        regularization_strength=regularization_strength,
                        l1_ratio=l1_ratio,
                    )
        # Add insurance problems
        for distribution in insurance_distributions:
            suffix = penalty_str + "-" + distribution
            dist = distribution
            for problem_name, load_fn in insurance_load_funcs.items():
                for data_setup in ["weights", "no-weights", "offset"]:
                    problems["-".join((problem_name, data_setup, suffix))] = Problem(
                        data_loader=partial(
                            load_data, load_fn, distribution=dist, data_setup=data_setup
                        ),
                        distribution=distribution,
                        regularization_strength=regularization_strength,
                        l1_ratio=l1_ratio,
                    )

    return problems
